
from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File containing commonly used similarity measures in medical image analysis. ",
    help = "Use `metric['ABBR.'](A, B)` to compute the similarity.",
    requires = "SimpleITK"
).check()

__all__ = """
    metric
    ITKMetric
    ITKLabelMetric
    MutualInformation
    NormalizedMutualInformation
    KLDivergence
    CorrelationOfLocalEstimation
    NormalizedVectorInformation
    Cos2Theta
    SumSquaredDifference
    MeanSquaredErrors
    PeakSignalToNoiseRatio
    CrossEntropy
    CrossCorrelation
    NormalizedCrossCorrelation
    StructuralSimilarity
    Dice DiceScore DiceScoreCoefficient
    LabelDice LabelDiceScore LabelDiceScoreCoefficient
    ITKDiceScore
    ITKJaccardCoefficient
    ITKVolumeSimilarity
    ITKFalsePositive
    ITKFalseNegative
    ITKHausdorffDistance
    ITKMedianSurfaceDistance
    ITKAverageSurfaceDistance
    ITKDivergenceOfSurfaceDistance
    ITKLabelDiceScore
    ITKLabelJaccardCoefficient
    ITKLabelVolumeSimilarity
    ITKLabelFalsePositive
    ITKLabelFalseNegative
    ITKLabelHausdorffDistance
    ITKLabelMedianSurfaceDistance
    ITKLabelAverageSurfaceDistance
    ITKLabelDivergenceOfSurfaceDistance
    LocalNonOrthogonality
    RigidProjectionError
    
    joint_hist
""".split()

with __info__:
    import torch
    import batorch as bt
    import numpy as np
    import SimpleITK as sitk
    from pycamia import to_tuple, avouch, Version, alias

######### Section 1: Information Based ########
eps = 1e-6

@bt.batorch_wrapper
def Bspline(i: bt.Tensor, U: bt.Tensor):
    """
    Cubic B-spline function. 
    Args: Segment i with argument U; i and U are of the same size. 
    """
    return (
        bt.where(i == -1, (1 - U) ** 3 / 6,
        bt.where(i == 0, U ** 3 / 2 - U * U + 2 / 3,
        bt.where(i == 1, (- 3 * U ** 3 + 3 * U * U + 3 * U + 1) / 6,
        bt.where(i == 2, U ** 3 / 6,
        bt.zeros_like(U)))))
    )

@bt.batorch_wrapper
def dBspline(i: bt.Tensor, U: bt.Tensor):
    """
    Derivative of cubic B-spline function. 
    Args: Segment i with argument U; i and U are of the same size. 
    """
    return (
        bt.where(i == -1, - 3 * (1 - U) ** 2 / 6,
        bt.where(i == 0, 3 * U ** 2 / 2 - 2 * U,
        bt.where(i == 1, (- 3 * U ** 2 + 2 * U + 1) / 2,
        bt.where(i == 2, 3 * U ** 2 / 6,
        bt.zeros_like(U)))))
    )

@bt.batorch_wrapper
def multi_Bspline(i: bt.Tensor, U: bt.Tensor):
    """
    Multi-variate cubic B-spline function. 
    Args: Segment i with argument U; i and U are of the same size:
        ({n_batch}, [n_image], n_1, ..., n_r) [r=n_dim] 
    """
    return Bspline(i, U).prod([])

@bt.batorch_wrapper
def multi_dBspline(i: bt.Tensor, U: bt.Tensor, j):
    """
    Derivative of j-th term of multi-variate cubic B-spline function. 
    Args: Segment i with argument U; i and U are of the same size:
        ({n_batch}, [n_image], n_1, ..., n_r) [r=n_dim] 
    """
    return Bspline(i, U)[:, :j].prod([]) * dBspline(i, U)[:, j] * Bspline(i, U)[:, j+1:].prod([])

class JointHistogram(bt.autograd.Function):
    
    @staticmethod
    @bt.batorch_wrapper
    def forward(ctx, I1: bt.Tensor, I2: bt.Tensor, nbin=100, mask=None):
        """batorch version of JointHistogram.
        """
        if mask is None: mask = bt.ones_like(I1)
        with bt.no_grad():
            n_dim = I1.n_space_dim
            n_bin = nbin
            data_pair = bt.stack(I1.flatten(1), I2.flatten(1), dim=[]) # ({n_batch}, [n_image=2], n_data)
            n_batch, n_image, n_data = tuple(data_pair.shape)
            device = data_pair.device
            indices = []; values = []
            l_window = 4
            window = (bt.image_grid([l_window] * n_image) - 1).flatten(...).transpose(0, 1).to(device) # (l_window ^ n_image, [n_image])
            for shift in window: # ([n_image],)
                hist_pos = data_pair * n_bin # ({n_batch}, [n_image], n_data)
                index = bt.clamp(bt.floor(hist_pos).long() + shift, 0, n_bin - 1)
                batch_idx = bt.arange(n_batch, device=device).view({n_batch}, [1], 1).repeat(1, 1, n_data)
                index = bt.cat(batch_idx, index, []) # ({n_batch}, [n_image + 1], n_data)
                value = multi_Bspline(shift.expand_to(data_pair), bt.decimal(hist_pos)) # ({n_batch}, n_data)
                value *= mask.flatten(...) # MASK HERE: ({n_batch}, n_data)
                indices.append(index)
                values.append(value)
            # ({n_batch}, [1 + n_image], n_data x l_window ^ n_image)
            Mindices = bt.cat(indices, -1)
            # ({n_batch}, n_data x l_window ^ n_image)
            Mvalues = bt.cat(values, -1)
            # ([1 + n_image], n_batch x n_data x l_window ^ n_image)
            indices = Mindices.merge_dims({}, -1)
            # (n_batch x n_data x l_window ^ n_image,)
            values = Mvalues.flatten(0)
            if Version(bt.__version__) >= Version("2"):
                creator = lambda *x: bt.torch.sparse_coo_tensor(*x, device=indices.device)
            elif indices.device.type == 'cpu': creator = bt.torch.sparse.FloatTensor
            else: creator = bt.torch.cuda.sparse.FloatTensor
            collected = creator(indices, values, (n_batch,) + (n_bin,) * n_image).to_dense()
            collected = bt.Tensor(collected, batch_dim=0) / mask.sum(...) # ({n_batch}, n_bin, ..., n_bin)

            # JH = collected / n_data
            JH = collected
            ctx.save_for_backward(bt.tensor(list(I1.shape)), data_pair, JH, mask)
        return collected

    @staticmethod
    def backward(ctx, grad_output):
        # grad_output: ({n_batch}, n_bin, ..., n_bin)
        grad_output = grad_output.as_subclass(bt.Tensor).init_special().with_batch_dim(True)
        Ishape, data_pair, _, mask = ctx.saved_tensors
        device = grad_output.device
        n_bin = grad_output.size(-1)
        Ishape = tuple(Ishape.int().tolist())
        with bt.no_grad():
            # data_pair: ({n_batch}, [n_image=2], n_data)
            n_batch, n_image, n_data = data_pair.shape
            dEdI1 = bt.zeros({n_batch}, n_data).to(device) # ({n_batch}, n_1, ..., n_r) [r=n_dim]
            dEdI2 = bt.zeros({n_batch}, n_data).to(device)
            if Version(bt.__version__) >= '1.9': window = bt.stack(bt.meshgrid(*([bt.arange(-1, 3)] * n_image), indexing='ij'), 0).flatten(1).transpose(0, 1).to(device) # (4 ^ n_image, n_image)
            else: window = bt.stack(bt.meshgrid(*([bt.arange(-1, 3)] * n_image), indexing='ij'), 0).flatten(1).transpose(0, 1).to(device) # (4 ^ n_image, n_image)
            for shift in window:
                shift = shift.view({1}, [2], 1) # ({1}, [n_image], 1) -> (1, n_image, 1)
                hist_pos = data_pair * n_bin # ({n_batch}, [n_image], n_data) -> (n_batch, n_image, n_data)
                index = bt.clamp(bt.floor(hist_pos).long() + shift.long(), 0, n_bin - 1)
                decimal = hist_pos - bt.floor(hist_pos)
                value = grad_output[(bt.arange(n_batch).long().unsqueeze(-1).expand(n_batch, n_data).to(device),) + tuple(x.squeeze(1) for x in index.split(1, 1))] # ({n_batch}, n_data) -> (n_batch, n_data)
                dEdI1 += value * multi_dBspline(shift, decimal, 0) # ({n_batch}, n_data) -> (n_batch, n_data)
                dEdI2 += value * multi_dBspline(shift, decimal, 1)
            # dEdI1 = bt.tensor(dEdI1.data).view(Ishape) * n_bin / n_data
            # dEdI2 = bt.tensor(dEdI2.data).view(Ishape) * n_bin / n_data
            n_dim = len(Ishape) - 1
            dEdI1 = dEdI1.data.view(Ishape) * mask * n_bin / mask.flatten(1).sum(1).view((n_batch,) + (1,) * n_dim)
            dEdI2 = dEdI2.data.view(Ishape) * mask * n_bin / mask.flatten(1).sum(1).view((n_batch,) + (1,) * n_dim)
        return dEdI1, dEdI2, None, None
        # Deprecated: using gather to pick elements is not necessary. 
        #         grad_y = grad_output[(slice(None),) + index.split(1, 1)].squeeze(2)
        #         value = grad_y.gather(0, bt.arange(nbatch).long().unsqueeze(0).unsqueeze(-1).repeat(1, 1, ndata)).view(ctx.Ishape)
        #         dPdI1 += value * multi_dBspline(shift, bt.decimal(data_pair * nbin), 0).view(ctx.Ishape)
        #         dPdI2 += value * multi_dBspline(shift, bt.decimal(data_pair * nbin), 1).view(ctx.Ishape)
        # return dPdI1, dPdI2, None

# class JointHistogram(bt.autograd.Function):
    
#     @staticmethod
#     def forward(ctx, I1, I2, nbin=100, mask=None):
#         """
#         Estimate the joint histogram between I1 & I2. 
        
#         Args:
#             I1 (bt.Tensor): (n_batch, n@1, n@2, ..., n@n_dim)
#             I2 (bt.Tensor): (n_batch, n@1, n@2, ..., n@n_dim)
#             nbin (int): the size of histogram. Note that it should be set to smaller numbers if 
#                 the number of intensity is low. Otherwise, gradients would be too small.

#         Returns:
#             histogram (bt.Tensor): (n_batch, n_bin, ..., n_bin)
#         """
#         if isinstance(I1, bt.Tensor): I1 = I1.as_subclass(bt.Tensor)
#         if isinstance(I2, bt.Tensor): I2 = I2.as_subclass(bt.Tensor)
#         if isinstance(mask, bt.Tensor): mask = mask.as_subclass(bt.Tensor)
#         device = I1.device
#         if mask is None: mask = bt.ones_like(I1).to(device)
#         with bt.no_grad():
#             if hasattr(ctx, 'JH'): del ctx.JH
#             n_bin = bt.tensor(nbin).to(device)
#             n_dim = I1.ndim - 1
#             # mask_pos = bt.stack(bt.meshgrid(*[bt.arange(x) for x in mask.shape[1:]]), 0).unsqueeze(0).to(device) * mask.unsqueeze(1) # (n_batch, n_dim, n@1, ..., n@n_dim)
#             # level = bt.tensor(bt.cat((bt.tensor([1]), bt.cumprod(bt.tensor(mask.shape[:0:-1]), 0)[:-1]), 0).tolist()[::-1]).to(device)
#             # mask_select = bt.unique((mask_pos * level.view((1, n_dim) + (1,) * n_dim)).sum(1)).long()
#             # data_pair = bt.stack((I1.flatten(1), I2.flatten(1)), 1)[..., mask_select] # ({n_batch}, {n_hist=2}, n_data) -> (n_batch, n_hist, n_data)
#             data_pair = bt.stack((I1.flatten(1), I2.flatten(1)), 1) # ({n_batch}, {n_hist=2}, n_data) -> (n_batch, n_hist, n_data)
#             n_batch, n_hist, n_data = data_pair.shape
#             indices = []; values = []
#             if Version(bt.__version__) >= '1.9': window = bt.stack(bt.meshgrid(*([bt.arange(-1, 3)] * n_hist), indexing='ij'), 0).flatten(1).transpose(0, 1).to(device) # (4 ^ n_hist, n_hist)
#             else: window = bt.stack(bt.meshgrid(*([bt.arange(-1, 3)] * n_hist)), 0).flatten(1).transpose(0, 1).to(device) # (4 ^ n_hist, n_hist)
#             for shift in window: # (n_hist,)
#                 hist_pos = data_pair * n_bin # ({n_batch}, {n_hist}, n_data) -> (n_batch, n_hist, n_data)
#                 decimal = hist_pos - bt.floor(hist_pos)
#                 index = bt.clamp(bt.floor(hist_pos).long() + shift.view(1, -1, 1).long(), 0, n_bin - 1)
#                 batch_idx = bt.arange(n_batch).unsqueeze(-1).unsqueeze(-1).expand(n_batch, 1, n_data).to(device)
#                 index = bt.cat((batch_idx, index), 1) # ({n_batch}, {n_hist + 1}, n_data) -> (n_batch, n_hist + 1, n_data)
#                 value = Bspline(shift.view(1, -1, 1), decimal).prod(1) # ({n_batch}, n_data)
#                 # value *= mask.flatten(1) # MASK HERE: ({n_batch}, n_data)
#                 indices.append(index)
#                 values.append(value)
#             # ({n_batch}, {1 + n_hist}, n_data x 4 ^ n_hist)
#             Mindices = bt.cat(indices, -1)
#             # ({n_batch}, n_data x 4 ^ n_hist)
#             Mvalues = bt.cat(values, -1)
#             # ({1 + n_hist}, n_batch x n_data x 4 ^ n_hist)
#             indices = Mindices.transpose(0, 1).flatten(1)
#             # (n_batch x n_data x 4 ^ n_hist,)
#             values = Mvalues.flatten(0)
#             if indices.device == bt.device('cpu'): creator = bt.sparse.FloatTensor
#             else: creator = bt.cuda.sparse.FloatTensor
#             collected = creator(indices, values, (n_batch,) + (n_bin,) * n_hist).to_dense()
#             # collected = bt.Tensor(collected, batch_dim=0) # ({n_batch}, n_bin, ..., n_bin)

#             ctx.nbin = n_bin
#             ctx.Ishape = I1.shape
#             ctx.data_pair = data_pair
#             # ctx.JH = collected / mask.flatten(1).sum(1).unsqueeze(-1).unsqueeze(-1)
#             ctx.JH = collected / n_data
#             ctx.mask = mask
#         return ctx.JH

#     @staticmethod
#     def backward(ctx, grad_output):
#         # grad_output: (n_batch, n_bin, ..., n_bin)
#         device = grad_output.device
#         with bt.no_grad():
#             n_bin = ctx.nbin # (scalar)
#             data_pair = ctx.data_pair # ({n_batch}, {n_hist=2}, n_data) -> (n_batch, n_hist=2, n_data)
#             n_batch, n_hist, n_data = data_pair.shape
#             dEdI1 = bt.zeros(n_batch, n_data).to(device) # ({n_batch}, n@1, ..., n@n_dim) -> (n_batch, n@1, ..., n@n_dim)
#             dEdI2 = bt.zeros(n_batch, n_data).to(device)
#             if Version(bt.__version__) >= '1.9': window = bt.stack(bt.meshgrid(*([bt.arange(-1, 3)] * n_hist), indexing='ij'), 0).flatten(1).transpose(0, 1).to(device) # (4 ^ n_hist, n_hist)
#             else: window = bt.stack(bt.meshgrid(*([bt.arange(-1, 3)] * n_hist)), 0).flatten(1).transpose(0, 1).to(device) # (4 ^ n_hist, n_hist)
#             for shift in window:
#                 shift = shift.view(1, 2, 1) # ([1], {n_hist}, 1) -> (1, n_hist, 1)
#                 hist_pos = data_pair * n_bin # ({n_batch}, {n_hist}, n_data) -> (n_batch, n_hist, n_data)
#                 index = bt.clamp(bt.floor(hist_pos).long() + shift.long(), 0, n_bin - 1)
#                 decimal = hist_pos - bt.floor(hist_pos)
#                 value = grad_output[(bt.arange(n_batch).long().unsqueeze(-1).expand(n_batch, n_data).to(device),) + tuple(x.squeeze(1) for x in index.split(1, 1))] # ({n_batch}, n_data) -> (n_batch, n_data)
#                 dEdI1 += value * multi_dBspline(shift, decimal, 0) # ({n_batch}, n_data) -> (n_batch, n_data)
#                 dEdI2 += value * multi_dBspline(shift, decimal, 1)
#             n_dim = len(ctx.Ishape) - 1
#             # dEdI1 = dEdI1.data.view(ctx.Ishape) * ctx.mask * n_bin / ctx.mask.flatten(1).sum(1).view((n_batch,) + (1,) * n_dim)
#             # dEdI2 = dEdI2.data.view(ctx.Ishape) * ctx.mask * n_bin / ctx.mask.flatten(1).sum(1).view((n_batch,) + (1,) * n_dim)
#             dEdI1 = dEdI1.data.view(ctx.Ishape) * n_bin / n_data
#             dEdI2 = dEdI2.data.view(ctx.Ishape) * n_bin / n_data
#             # n_data = 1
#             # for x in ctx.Ishape[1:]: n_data *= x
#             # dEdI1_whole = bt.zeros(n_batch, n_data).to(device)
#             # dEdI1_whole[..., ctx.mask_select] = dEdI1
#             # dEdI1 = dEdI1_whole.data.view(ctx.Ishape) * n_bin / n_data
#             # dEdI2_whole = bt.zeros(n_batch, n_data).to(device)
#             # dEdI2_whole[..., ctx.mask_select] = dEdI2
#             # dEdI2 = dEdI2_whole.data.view(ctx.Ishape) * n_bin / n_data
#         return dEdI1, dEdI2, None, None

def joint_hist(A, B, nbin=100, mask=None):
    if not A.has_batch: A = A.unsqueeze({})
    if not B.has_batch: B = B.unsqueeze({})
    return JointHistogram.apply(A, B, nbin, mask)

def MutualInformation(A, B, nbin=100, mask=None):
    func = 'MutualInformation'
    n_batch, i_batch, i_channel = None, None, None
    if isinstance(A, bt.Tensor):
        if A.has_channel and A.has_batch: i_channel = A.channel_dim; A = A.mergedims([], {})
        if A.has_channel and not A.has_batch: i_channel = A.channel_dim; A.with_channeldim(None).with_batchdim(i_channel)
        if A.has_batch: n_batch = A.n_batch; i_batch = A.batch_dim; A = A.movedim({}, 0)
        # A = A.as_subclass(bt.Tensor)
    if isinstance(B, bt.Tensor):
        if B.has_channel and B.has_batch: B = B.mergedims([], {})
        if B.has_channel and not B.has_batch: B.with_channeldim(None).with_batchdim(i_channel)
        if B.has_batch: B = B.movedim({}, 0)
        # B = B.as_subclass(bt.Tensor)
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape: {A.shape} and {B.shape}")
    avouch(A.max() <= 1 and A.min() >= 0 and B.max() <= 1 and B.min() >= 0, f"Please make sure inputs of '{func}' are normalized images with intensity in [0, 1]. Currently of ranges [{A.min().item()}, {A.max().item()}] and [{B.min().item()}, {B.max().item()}].")

    Pab = JointHistogram.apply(A, B, nbin, mask)
    Pa = Pab.sum(2); Pb = Pab.sum(1)
    Hxy = - bt.sum(Pab * bt.log2(bt.where(Pab < eps, bt.ones_like(Pab), Pab)), [1, 2])
    Hx = - bt.sum(Pa * bt.log2(bt.where(Pa < eps, bt.ones_like(Pa), Pa)), 1)
    Hy = - bt.sum(Pb * bt.log2(bt.where(Pb < eps, bt.ones_like(Pb), Pb)), 1)
    MI = (Hx + Hy - Hxy).as_subclass(bt.Tensor).init_special()
    if i_batch is not None: MI.with_batchdim(0).movedim_(0, i_batch)
    if i_channel is not None and i_channel == i_batch: MI.with_batchdim(None).with_channeldim(i_channel)
    if i_channel is not None and i_batch is not None: MI = MI.splitdim({}, {n_batch}, []).movedim([], i_channel)
    return MI

def NormalizedMutualInformation(A, B, nbin=100, mask=None):
    func = 'NormalizedMutualInformation'
    n_batch, i_batch, i_channel = None, None, None
    if isinstance(A, bt.Tensor):
        if A.has_channel and A.has_batch: i_channel = A.channel_dim; A = A.mergedims([], {})
        if A.has_channel and not A.has_batch: i_channel = A.channel_dim; A.with_channeldim(None).with_batchdim(i_channel)
        if A.has_batch: n_batch = A.n_batch; i_batch = A.batch_dim; A = A.movedim({}, 0)
    if isinstance(B, bt.Tensor):
        if B.has_channel and B.has_batch: B = B.mergedims([], {})
        if B.has_channel and not B.has_batch: B.with_channeldim(None).with_batchdim(i_channel)
        if B.has_batch: B = B.movedim({}, 0)
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape: {A.shape} and {B.shape}")
    avouch(A.max() <= 1 + eps and A.min() >= -eps and B.max() <= 1 + eps and B.min() >= -eps, f"Please make sure inputs of '{func}' are normalized images with intensity in [0, 1]. Currently of ranges [{A.min().item()}, {A.max().item()}] and [{B.min().item()}, {B.max().item()}].")

    Pab = JointHistogram.apply(A, B, nbin, mask)
    # Pab = JointHistogram.apply(A, B, nbin, mask)
    Pa = Pab.sum(2); Pb = Pab.sum(1)
    Hxy = - bt.sum(Pab * bt.log2(bt.where(Pab < eps, bt.ones_like(Pab), Pab)), [1, 2])
    Hx = - bt.sum(Pa * bt.log2(bt.where(Pa < eps, bt.ones_like(Pa), Pa)), 1)
    Hy = - bt.sum(Pb * bt.log2(bt.where(Pb < eps, bt.ones_like(Pb), Pb)), 1)
    NMI = bt.divide(Hx + Hy, Hxy, 0.0)
    if i_batch is not None: NMI.movedim_({}, i_batch)
    if i_channel is not None and i_channel == i_batch: NMI.with_batchdim(None).with_channeldim(i_channel)
    if i_channel is not None and i_batch is not None: NMI = NMI.splitdim({}, {n_batch}, []).movedim([], i_channel)
    return NMI

def KLDivergence(A, B, nbin=100):
    func = 'KLDivergence'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")
    avouch(A.max() <= 1 and A.min() >= 0 and B.max() <= 1 and B.min() >= 0, f"Please make sure inputs of '{func}' are normalized images with intensity in [0, 1]. Currently of ranges [{A.min().item()}, {A.max().item()}] and [{B.min().item()}, {B.max().item()}].")

    Pab = JointHistogram_bt.apply(A, B, nbin)
    Pa = Pab.sum(2); Pb = Pab.sum(1)
    return (Pa * bt.log2(bt.where(Pb < eps, bt.ones_like(Pa), Pa / Pb.clamp(min=eps)).clamp(min=eps))).sum(1)

###############################################

######## Section 2: Cross Correlation #########

@bt.batorch_wrapper
def local_matrix(A: bt.Tensor, B: bt.Tensor, s=0, kernel="Gaussian", kernel_size=3):
    if isinstance(kernel, str):
        if kernel.lower() == "gaussian": kernel = bt.gaussian_kernel(n_dims = A.n_space_dim, kernel_size = kernel_size).unsqueeze(0, 0)
        elif kernel.lower() == "mean": kernel = bt.ones(*(kernel_size,) * A.n_space_dim).unsqueeze(0, 0) / (kernel_size ** A.n_space_dim)
    elif hasattr(kernel, 'shape'): kernel_size = kernel.size(-1)

    def mean(a):
        op = eval("bt.nn.functional.conv%dd"%A.n_space_dim)
        if a.has_batch: x = a.unsqueeze([])
        else: x = a.unsqueeze({}, [])
        return op(x, kernel, padding = kernel_size // 2).squeeze(*((1,) if a.has_batch else (0, 0)))

    if s > 0:
        GA = bt.grad_image(A, pad=True)
        GB = bt.grad_image(B, pad=True)
        point_estim = bt.stack(bt.dot(GA, GA), bt.dot(GA, GB), bt.dot(GB, GB), dim={int(A.has_batch)})
    else: point_estim = 0

    MA = mean(A)
    MB = mean(B)
    local_estim = bt.stack(mean(A * A) - MA ** 2, mean(A * B) - MA * MB, mean(B * B) - MB ** 2, dim={int(A.has_batch)})

    return s * point_estim + local_estim

def CorrelationOfLocalEstimation(A, B, s=0, kernel="Gaussian", kernel_size=3):
    func = 'CorrelationOfLocalEstimation'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    S11, S12, S22 = local_matrix(A, B, s=s, kernel=kernel, kernel_size=kernel_size).split()
    return (bt.divide(S12 ** 2, S11 * S22, tol=eps).squeeze(1) + eps).sqrt().mean(...)

###############################################

########## Section 3: Local Gradient ##########

def NormalizedVectorInformation(A, B, mask=None):
    func = 'NormalizedVectorInformation'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    GA = bt.grad_image(A, pad=True)
    GB = bt.grad_image(B, pad=True)
    if mask is None: mask = bt.ones_like(A)
    return (bt.divide(bt.dot(GA, GB) ** 2, bt.dot(GA, GA) * bt.dot(GB, GB), tol=eps) * mask).mean(...)

def Cos2Theta(A, B, mask=None):
    func = 'Cos2Theta'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    GA = bt.grad_image(A, pad=True)
    GB = bt.grad_image(B, pad=True)
    if mask is None: mask = bt.ones_like(A)
    return (bt.divide(bt.dot(GA, GB) ** 2, bt.dot(GA, GA) * bt.dot(GB, GB), tol=eps) * mask)

###############################################

####### Section 4: Intensity Difference #######

@alias("SumOfSquaredDifference")
def SumSquaredDifference(A, B, mask=None):
    func = 'SumSquaredDifference'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if mask is None: mask = bt.ones_like(A)
    return ((A - B) ** 2 * mask).sum(...)

@alias("MeanOfSquaredErrors")
def MeanSquaredErrors(A, B, mask=None):
    func = 'MeanSquaredErrors'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if mask is None: mask = bt.ones_like(A)
    return ((A - B) ** 2 * mask).mean(...)

def PeakSignalToNoiseRatio(A, B, mask=None):
    func = 'PeakSignalToNoiseRatio'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if mask is None: mask = bt.ones_like(A)
    return 10 * bt.log10(bt.max(((A * mask).max(), (B * mask).max())) ** 2 / ((A - B) ** 2 * mask).mean(...))

###############################################

##### Section 5: Distribution Similarity ######

def CrossEntropy(y, label):
    func = 'CrossEntropy'
    avouch(isinstance(y, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(label, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(y.has_batch and label.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(y.has_channel and label.has_channel, f"Please make sure inputs of '{func}' have channel dimensions to calculate entropy along." +
           "Use X.channel_dim = 0 to identify (or X.unsqueeze([]) if no existed channel, though this should not be commonly seen).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    ce = - label * bt.log(y.clamp(1e-10, 1.0))
    return ce.sum(ce.channel_dimension).mean(...)

def CrossCorrelation(A, B, mask=None):
    func = 'CrossCorrelation'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if mask is None: mask = bt.ones_like(A)
    dA = A - (A * mask).sum(...) / mask.sum(...); dB = B - (B * mask).sum(...) / mask.sum(...)
    return (dA * dB * mask).sum(...)

def NormalizedCrossCorrelation(A, B, mask=None):
    func = 'NormalizedCrossCorrelation'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if mask is None: mask = bt.ones_like(A)
    dA = A - (A * mask).sum(...) / mask.sum(...); dB = B - (B * mask).sum(...) / mask.sum(...)
    return (dA * dB * mask).sum(...) / (dA ** 2 * mask).sum(...).sqrt() / (dB ** 2 * mask).sum(...).sqrt()

def StructuralSimilarity(A, B, k1=0.01, k2=0.03, mask=None):
    func = 'StructuralSimilarity'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if mask is None: mask = bt.ones_like(A)
    A_mean = (A * mask).sum(...) / mask.sum(...)
    B_mean = (B * mask).sum(...) / mask.sum(...)
    dA = A - A_mean; dB = B - B_mean
    varA = (dA ** 2 * mask).mean(...)
    varB = (dB ** 2 * mask).mean(...)
    covAB = (dA * dB * mask).mean(...)
    L = bt.max(((A * mask).max(), (B * mask).max()))
    c1, c2 = k1 * L, k2 * L
    num = (2 * A_mean * B_mean + c1 ** 2) * (2 * covAB + c2 ** 2)
    den = (A_mean ** 2 + B_mean ** 2 + c1 ** 2) * (varA + varB + c2 ** 2)
    return num / den

###############################################

########## Section 6: Region Overlap ##########

@alias('Dice', 'DiceScore')
def DiceScoreCoefficient(A, B, mask=None):
    '''
    The Dice score between A and B where A and B are 0-1 masks. 
    The sizes are as follows: 
    A: ({n_batch}, [n_label: optional], n@1, n@2, ..., n@n_dim)
    B: ({n_batch}, [n_label: optional], n@1, n@2, ..., n@n_dim)
    return: ({n_batch}, [n_label: optional])
    '''
    func = 'Dice'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if mask is None: mask = bt.ones_like(A)
    ABsum = (A * mask).sum(...) + (B * mask).sum(...)
    return 2 * (A * B * mask).sum(...) / (ABsum + eps)

@alias('LabelDice', 'LabelDiceScore')
def LabelDiceScoreCoefficient(A, B, class_labels=None, mask=None):
    '''
    The Dice score between A and B where A and B are integer label maps. 
    
    Params:
        A [bt.Tensor]: label map 1 with size ({n_batch}, n@1, ..., n@n_dim).
        B [bt.Tensor]: label map 2 with size ({n_batch}, n@1, ..., n@n_dim).
        class_labels [list or NoneType]: integers representing different labels, a list of length `n_class`. 
            If it is not given, it will be automatically detected by collecting all sorted labels (except the minimum as background) in A and B. 
            It is time consuming, especially if A and B are accidentally float images. Please be careful when using this default. 
        
    output [bt.Tensor]: the Dice scores for each label. 
        size: ({n_batch}, {n_class})
    '''
    func = 'LabelDice'
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if not class_labels: class_labels = sorted(list(set(A.unique().tolist() + B.unique().tolist())))[1:]
    A_labels = [1 - bt.clamp(bt.abs(A - i), 0, 1) for i in class_labels]
    B_labels = [1 - bt.clamp(bt.abs(B - i), 0, 1) for i in class_labels]
    A_maps = bt.stack(A_labels, [1])
    B_maps = bt.stack(B_labels, [1])
    return Dice(A_maps, B_maps, mask=mask)

###############################################

######### Section 7: Surface distance #########
class SurfaceDistanceImageFilter:
    def __init__(self): self.all_dis = bt.tensor([0])
    def Execute(self, A, B):
        array = lambda x: np.array(sitk.GetArrayViewFromImage(x)).astype(np.float32)
        ADisMap = sitk.Abs(sitk.SignedMaurerDistanceMap(A, squaredDistance = False, useImageSpacing = True))
        BDisMap = sitk.Abs(sitk.SignedMaurerDistanceMap(B, squaredDistance = False, useImageSpacing = True))
        Asurface = sitk.LabelContour(A)
        Bsurface = sitk.LabelContour(B)
        
        # for a pixel 'a' in A, compute aBdis = dis(a, B)
        aBDis = array(BDisMap)[array(Asurface) > 0]
        # for a pixel 'b' in B, compute aBdis = dis(b, A)
        bADis = array(ADisMap)[array(Bsurface) > 0]
        if aBDis.size == 0 or bADis.size == 0:
            raise TypeError("Cannot detect surfaces. ")
        self.all_dis = bt.tensor(np.concatenate((aBDis, bADis), 0))
        
    def GetHausdorffDistance(self): return self.all_dis.max()
    def GetMedianSurfaceDistance(self): return self.all_dis.median()
    def GetAverageSurfaceDistance(self): return self.all_dis.mean()
    def GetDivergenceOfSurfaceDistance(self): return self.all_dis.std()

def ITKMetric(A, B, spacing = 1, metric = "HD"):
    '''
    The metrics between A and B where A and B are 0-1 masks. 
    The sizes are as follows: 
    A: ({n_batch}, [n_label: optional], n@1, n@2, ..., n@n_dim)
    B: ({n_batch}, [n_label: optional], n@1, n@2, ..., n@n_dim)
    return: ({n_batch}, [n_label: optional])
    '''
    func = 'Metric ' + metric
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    n_batch = A.n_batch
    has_channel = False
    if A.has_channel and B.has_channel:
        has_channel = True
        n_channel = A.n_channel
        A = A.mergedims([], {})
        B = B.mergedims([], {})
    n_maps = A.n_batch
    filter_A = A.sum(...) != 0
    filter_B = B.sum(...) != 0
    filter_both = filter_A & filter_B
    A = A[filter_both]
    B = B[filter_both]
    A = np.array(A) != 0
    B = np.array(B) != 0
    spacing = to_tuple(spacing)
    n_dim = A.ndim
    n_data = A.shape[0]
    if len(spacing) == 1: spacing *= n_dim
    Overlap_filter = sitk.LabelOverlapMeasuresImageFilter()
    SD_filter = SurfaceDistanceImageFilter()
    Overlap_execs = {
        'Dice': lambda x: x.GetDiceCoefficient(),
        'Jaccard': lambda x: x.GetJaccardCoefficient(),
        'Volume': lambda x: x.GetVolumeSimilarity(),
        'Falsepositive': lambda x: x.GetFalsePositiveError(),
        'Falsenegative': lambda x: x.GetFalseNegativeError()
    }
    SD_execs = {
        'HD': lambda x: x.GetHausdorffDistance(),
        'MSD': lambda x: x.GetMedianSurfaceDistance(),
        'ASD': lambda x: x.GetAverageSurfaceDistance(),
        'STDSD': lambda x: x.GetDivergenceOfSurfaceDistance()
    }
    measures = np.zeros((n_data,))
    for b in range(n_data):
        ITKA = sitk.GetImageFromArray(A[b].astype(np.int32), isVector = False)
        ITKA.SetSpacing(spacing)
        ITKB = sitk.GetImageFromArray(B[b].astype(np.int32), isVector = False)
        ITKB.SetSpacing(spacing)
        metric = metric.capitalize()
        if metric in Overlap_execs:
            Overlap_filter.Execute(ITKA, ITKB)
            measures[b] = Overlap_execs[metric](Overlap_filter)
        metric = metric.upper()
        if metric in SD_execs:
            SD_filter.Execute(ITKA, ITKB)
            measures[b] = SD_execs[metric](SD_filter)
    
    resulting_measures = bt.zeros({n_maps})
    resulting_measures[filter_both] = bt.tensor_like(measures, target=resulting_measures)
    if has_channel:
        return resulting_measures.view({n_batch}, [n_channel])
    return resulting_measures.batch_dimension_(0)

def ITKLabelMetric(A, B, spacing = 1, metric = "HD", class_labels = None):
    '''
    The metrics between A and B where A and B are integer label maps. 
    
    Params:
        A [bt.Tensor]: label map 1 with size ({n_batch}, n@1, ..., n@n_dim).
        B [bt.Tensor]: label map 2 with size ({n_batch}, n@1, ..., n@n_dim).
        class_labels [list or NoneType]: integers representing different labels, a list of length `n_class`. 
            If it is not given, it will be automatically detected by collecting all sorted labels (except the minimum as background) in A and B. 
            It is time consuming, especially if A and B are accidentally float images. Please be careful when using this default. 
        
    output [bt.Tensor]: the metric values for each label. 
        size: ({n_batch}, [n_class])
    '''
    func = 'LabelMetric ' + metric
    avouch(isinstance(A, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(isinstance(B, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(A.has_batch and B.has_batch, f"Please make sure inputs of '{func}' have batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(A.shape == B.shape, f"Please make sure inputs of '{func}' have the same shape.")

    if not class_labels: class_labels = sorted(list(set(A.unique().tolist() + B.unique().tolist())))[1:]
    n_batch = A.n_batch
    n_class = len(class_labels)
    A_labels = [A == i for i in class_labels]
    B_labels = [B == i for i in class_labels]
    A_maps = bt.stack(A_labels, [])
    B_maps = bt.stack(B_labels, [])
    metric = ITKMetric(A_maps, B_maps, spacing, metric)
    return metric.view([n_class], {n_batch}).transpose(0, 1)

def ITKLabelDiceScore(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='Dice')
def ITKLabelJaccardCoefficient(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='Jaccard')
def ITKLabelVolumeSimilarity(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='Volume')
def ITKLabelFalsePositive(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='FalsePositive')
def ITKLabelFalseNegative(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='FalseNegative')
def ITKLabelHausdorffDistance(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='HD')
def ITKLabelMedianSurfaceDistance(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='MSD')
def ITKLabelAverageSurfaceDistance(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='ASD')
def ITKLabelDivergenceOfSurfaceDistance(A, B, spacing = 1, class_labels=None):
    return ITKLabelMetric(A, B, spacing=spacing, class_labels=class_labels, metric='STDSD')

def ITKDiceScore(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='Dice')
def ITKJaccardCoefficient(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='Jaccard')
def ITKVolumeSimilarity(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='Volume')
def ITKFalsePositive(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='FalsePositive')
def ITKFalseNegative(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='FalseNegative')
def ITKHausdorffDistance(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='HD')
def ITKMedianSurfaceDistance(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='MSD')
def ITKAverageSurfaceDistance(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='ASD')
def ITKDivergenceOfSurfaceDistance(A, B, spacing = 1):
    return ITKMetric(A, B, spacing=spacing, metric='STDSD')

###############################################

########## Section 6: Region Overlap ##########

def LocalNonOrthogonality(Disp_world_coord, mask = None, target_affine=None):
    """
    Local non-orthogonality metric defined for a displacement Disp_world_coord [?].
        Please use trans.toDDF(shape) to convert a transformation into a displacement field first before using the function. 
        Use .to_world_space to transform it to world space if necessary.
        Please refer to the reference [?] for more information. 
    [?] To be added. 

    Args:
        Disp_world_coord (bt.Tensor): The displacements for calculation. 
            size: ({n_batch}, {n_dim}, n@1, ..., n@n_dim)
        mask (bt.Tensor or NoneType): The mask in which we calculate the metric. It is the whole image by default. 
            size: ({n_batch}, n@1, ..., n@n_dim)
        spacing (int or tuple): The spacing of space

    Returns:
        bt.Tensor: Values of size ({n_batch},)
    """
    from .funcs import dilate
    func = 'LocalNonOrthogonality'
    avouch(isinstance(Disp_world_coord, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(mask is None or isinstance(mask, bt.Tensor), f"Please use 'babt.Tensor' object 'mask' for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(Disp_world_coord.has_batch, f"Please make sure inputs of '{func}' has batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(Disp_world_coord.has_channel, f"Please make sure input displacement of '{func}' has a channel dimension for the coordinates. Use X.channel_dim = 0 to identify. ")
    # spacing = to_tuple(spacing)
    # if len(spacing) < Disp_world_coord.n_space_dim: spacing *= Disp_world_coord.n_space_dim
    if target_affine is None: target_affine = lambda x: x

    n_batch = Disp_world_coord.n_batch
    if mask is None: mask = bt.ones(n_batch, *Disp_world_coord.space)
    mask = dilate((mask.float() > 0).float(), -1)
    X = target_affine(bt.image_grid(Disp_world_coord).duplicate(n_batch, {}).float()) # ({n_batch}, [n_dim], n@1, n@2, ..., n@n_dim)
    gd = bt.Jacobian(X, X + Disp_world_coord, dt=2) # of size ({n_batch}, n_dim, {n_dim}, n@1-dx, ..., n@n_dim-dx)
    JacOfPoints = gd.flatten(3).with_channeldim(None).mergedims(3, 0) # ([n_batch x n_data(-dx)], n_dim, n_dim)
    RigOfPoints = bt.Fnorm(JacOfPoints.T @ JacOfPoints - bt.eye_like(JacOfPoints)).view({n_batch}, -1) # ({n_batch}, n_data(-dx))
    MaskedRig = RigOfPoints.abs().view({n_batch}, *tuple(x-2 for x in Disp_world_coord.space)) * mask[(slice(None),) + (slice(1, -1),) * Disp_world_coord.n_channel]
    # print(JacOfPoints[MaskedRig[0].argmax()])
    # from micomputing import plot as plt
    # plt.subplots(3)
    # plt.gridshow(Disp_world_coord[0], on=bt.ones(Disp_world_coord.space), color='gray', as_orient="ILP")
    # plt.imshow(RigOfPoints.abs().view({n_batch}, *tuple(x-2 for x in Disp_world_coord.space))[0], as_orient = "ILP")
    # plt.imshow(MaskedRig[0], as_orient = "ILP")
    # plt.show()
    return MaskedRig.sum(...) / mask.sum(...)
    # return (RigOfPoints * bt.crop_as(mask, gd.shape[3:]).flatten(1)).sum(...) / mask.sum(...)
    
def RigidProjectionError(Disp_world_coord, mask = None, target_affine=None):
    '''
        Rigid projection error measures the rigidity of a matrix by evaluating the distance between it and its projection on rigid matrices. 
            This is defined for a displacement Disp_world_coord [?]. Please use trans.toDDF(shape) to convert a transformation into a displacement 
            field first before using the function. Use .to_world_space to transform it to world space if necessary.
            Please refer to the reference [?] for more information. 
        [?] To be added. 

        Disp_world_coord [bt.Tensor]: The displacements for calculation. 
            size: ({n_batch}, [n_dim], n@1, ..., n@n_dim)
        mask [bt.Tensor or NoneType]: The mask in which we calculate the metric. It is the whole image by default. 
            size: ({n_batch}, [n_region:optional], n@1, ..., n@n_dim)
        output [bt.Tensor]: Values of size ({n_batch},)
    '''
    func = 'LocalNonOrthogonality'
    avouch(isinstance(Disp_world_coord, bt.Tensor), f"Please use 'babt.Tensor' objects for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(mask is None or isinstance(mask, bt.Tensor), f"Please use 'babt.Tensor' object 'mask' for '{func}' in 'micomputing'. Use X.as_subclass(bt.Tensor) to create one. ")
    avouch(Disp_world_coord.has_batch and mask.has_batch, f"Please make sure inputs of '{func}' has batch dimensions. Use X.batch_dim = 0 to identify (or X.unsqueeze({{}}) if no existed batch).")
    avouch(Disp_world_coord.has_channel, f"Please make sure input displacement of '{func}' has a channel dimension for the coordinates. Use X.channel_dim = 0 to identify. ")
    
    def K(r):
        '''r: ({n_batch}, n_dim); K(r): ({n_batch}, n_dim, n_dim)'''
        return bt.cross_matrix(r)
    def Q(r):
        '''r: ({n_batch}, n_dim + 1); Q(r): ({n_batch}, n_dim + 1, n_dim + 1)'''
        r13, r4 = r.split([r.size(1) - 1, 1], 1)
        return bt.cat(bt.cat(r4 * bt.eye(r13) + K(r13), -r13.unsqueeze(1), 1), r.unsqueeze(-1), -1)
    def W(r):
        '''r: ({n_batch}, n_dim + 1); W(r): ({n_batch}, n_dim + 1, n_dim + 1)'''
        r13, r4 = r.split([r.size(1) - 1, 1], 1)
        return bt.cat(bt.cat(r4 * bt.eye(r13) - K(r13), -r13.unsqueeze(1), 1), r.unsqueeze(-1), -1)
    def R(r):
        '''r: ({n_batch}, n_dim + 1); R(r): ({n_batch}, n_dim, n_dim)'''
        n_dim = r.size(1) - 1
        return (W(r).T @ Q(r))[..., :n_dim, :n_dim]
    def t(r, s):
        '''r: ({n_batch}, n_dim + 1); s: ({n_batch}, n_dim + 1); t(r, s): ({n_batch}, n_dim)'''
        n_batch = r.n_batch
        n_dim = r.size(1) - 1
        return 2 * (W(r).T @ s)[..., :n_dim]
    def T(r, s):
        '''r: ({n_batch}, n_dim + 1); s: ({n_batch}, n_dim + 1); T(r, s): ({n_batch}, n_dim + 1, n_dim + 1)'''
        n_batch = r.n_batch
        n_dim = r.size(1) - 1
        return bt.cat(bt.cat(R(r), bt.zeros({n_batch}, 1, n_dim), 1), bt.cat(t(r, s).unsqueeze(-1), bt.ones({n_batch}, 1, 1), 1), -1)
    def max_eigvec(A):
        '''A: ({n_batch}, n_dim + 1, n_dim + 1); max_eigvec(A): ({n_batch}, n_dim + 1)'''
        n_batch = A.n_batch
        n_dim = A.size(1) - 1
        max_vec = bt.zeros({n_batch}, n_dim + 1)
        for b in range(n_batch):
            if Version(bt.bt.__version__) >= '1.10':
                l, v = bt.linalg.eig(A[b])
                max_vec[b] = bt.real(v)[:, bt.real(l).argmax()]
            else:
                l, v = bt.eig(A[b], eigenvectors=True)
                max_vec[b] = v[:, l[:, 0].argmax()]
        return max_vec
    def max_eigvec_iter(A):
        '''A: ({n_batch}, n_dim + 1, n_dim + 1); max_eigvec(A): ({n_batch}, n_dim + 1)'''
        n_batch = A.n_batch
        n_dim = A.size(1) - 1
        x = bt.ones({n_batch}, n_dim + 1)
        for _ in range(40): x += 2e-3 * ((x**2).sum(...) * (A @ x) - (x * (A @ x)).sum(...) * x) / (x**2).sum(...) ** 2
        return x / bt.norm(x) # eigenvalue: ((x * (A @ x)).sum(...) / (x**2).sum(...))
    
    n_batch = Disp_world_coord.n_batch
    if mask is None: mask = bt.ones({n_batch}, {1}, *Disp_world_coord.shape[2:])
    if mask.max() > 1:
        labels = sorted(mask.unique().tolist())[1:]
        mask = bt.stack([mask == l for l in labels], {})
    if not mask.has_channel:
        mask = mask.unsqueeze({})
    avouch(mask.has_channel, f"Please make sure mask of '{func}' has a channel dimension for the coordinates. Use X.channel_dim = 0 to identify. ")
    # spacing = to_tuple(spacing)
    # if len(spacing) < Disp_world_coord.n_space_dim: spacing *= Disp_world_coord.n_space_dim
    
    n_region = mask.n_channel
    X = bt.image_grid(Disp_world_coord).duplicate(n_batch, {}).float() # ({n_batch}, {n_dim}, n@1, n@2, ..., n@n_dim)
    if target_affine: X = target_affine(X)
    Y = X + Disp_world_coord
    if Disp_world_coord.n_channel == 2:
        X = bt.cat(X, bt.zeros({n_batch}, {1}), 1)
        Y = bt.cat(Y, bt.zeros({n_batch}, {1}), 1)
    n_dim = X.n_channel
    maxes = X.flatten(2).max(2).values.max({}).values # ({n_batch},)
    X = X * mask.unsqueeze({2})
    Y = Y * mask.unsqueeze({2}) # ({n_batch}, n_region, {n_dim}, n@1, n@2, ..., n@n_dim)
    p_mod = (X / maxes).flatten(3).mergedims(1, 3, 0).with_channeldim(None)
    p_obs = (Y / maxes).flatten(3).mergedims(1, 3, 0).with_channeldim(None) # ([n_batch x n_region x n_data], {n_dim})
    matW = W(bt.cat(p_mod, bt.zeros([p_mod.n_batch], 1), 1) / 2)
    matQ = Q(bt.cat(p_obs, bt.zeros([p_obs.n_batch], 1), 1) / 2) # ([n_batch x n_region x n_data], n_dim + 1, n_dim + 1)
    C1 = - 2 * matQ.T @ matW
    C1 = (C1.view({n_batch}, {n_region}, -1, n_dim + 1, n_dim + 1) * mask.flatten(2).unsqueeze(-1, -1)).sum(2) # ({n_batch}, {n_region}, n_dim + 1, n_dim + 1)
    if n_dim == 2: C1 -= 2 * bt.diag([-1, -1, 1, 1]).duplicate(n_batch, {}).duplicate(n_region, [])
    l = mask.sum(...).clamp(1)
    C2 = l * bt.eye(C1)
    C3 = 2 * ((matW - matQ).view({n_batch}, {n_region}, -1, n_dim + 1, n_dim + 1) * mask.flatten(2).unsqueeze(-1, -1)).sum(2)
    A = (C3.T @ C3 / (2 * l) - C1 - C1.T) / 2 # A = (C3.T @ (C2 + C2.T).inv() @ C3 - C1 - C1.T) / 2; it can be simplified by the substitution of C2
    r = max_eigvec(A.mergedims([], {}))
    s = - C3.mergedims([], {}) @ r / (2 * l.mergedims([], {})) # s = - (C2 + C2.T).inv() @ C3 @ r; it can be simplified by the substitution of C2
    matT = T(r, s).splitdim([], {n_batch}, {n_region})
    hatY = ((matT @ bt.cat(X.with_channeldim(2) / maxes, bt.ones({n_batch}, n_region, {1}), {}).flatten(3).with_channeldim(1))[:, :, :-1] * maxes).view({n_batch}, n_region, {n_dim}, *Disp_world_coord.shape[2:])
    return ((bt.Fnorm(Y - hatY).with_channeldim(1) * mask).sum(...) / l).mean({})

###############################################

# Metric abbreviations
class Metric:
    @alias("__getitem__")
    def __call__(self, key):
        """
            List
            ----------
            MI = MutualInformation,
            NMI = NormalizedMutualInformation,
            KL = KLDivergence,
            CLE = CorrelationOfLocalEstimation,
            NVI = NormalizedVectorInformation,
            SSD = SumSquaredDifference,
            MSE = MeanSquaredErrors,
            PSNR = PeakSignalToNoiseRatio,
            CE = CrossEntropy,
            CC = CrossCorrelation,
            NCC = NormalizedCrossCorrelation,
            SSIM = StructuralSimilarity,
            DSC = LabelDiceScore,
            JCD = ITKLabelJaccardCoefficient,
            VS = ITKLabelVolumeSimilarity,
            FP = ITKLabelFalsePositive,
            FN = ITKLabelFalseNegative,
            HD = ITKLabelHausdorffDistance,
            MdSD = ITKLabelMedianSurfaceDistance,
            ASD = ITKLabelAverageSurfaceDistance,
            MSD = ITKLabelAverageSurfaceDistance,
            divSD = ITKLabelDivergenceOfSurfaceDistance,
            stdSD = ITKLabelDivergenceOfSurfaceDistance,
            LNO = LocalNonOrthogonality,
            RPE = RigidProjectionError
        """
        return dict(
            MI = MutualInformation,
            NMI = NormalizedMutualInformation,
            KL = KLDivergence,
            CLE = CorrelationOfLocalEstimation,
            NVI = NormalizedVectorInformation,
            SSD = SumSquaredDifference,
            MSE = MeanSquaredErrors,
            PSNR = PeakSignalToNoiseRatio,
            CE = CrossEntropy,
            CC = CrossCorrelation,
            NCC = NormalizedCrossCorrelation,
            SSIM = StructuralSimilarity,
            DSC = LabelDiceScore,
            JCD = ITKLabelJaccardCoefficient,
            VS = ITKLabelVolumeSimilarity,
            FP = ITKLabelFalsePositive,
            FN = ITKLabelFalseNegative,
            HD = ITKLabelHausdorffDistance,
            MdSD = ITKLabelMedianSurfaceDistance,
            ASD = ITKLabelAverageSurfaceDistance,
            MSD = ITKLabelAverageSurfaceDistance,
            divSD = ITKLabelDivergenceOfSurfaceDistance,
            stdSD = ITKLabelDivergenceOfSurfaceDistance,
            LNO = LocalNonOrthogonality,
            RPE = RigidProjectionError
       )[key]
metric = Metric()




from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "micomputing",
    author = "Yuncheng Zhou",
    create = "2022-02",
    fileinfo = "File to read image data.",
    help = "Use `@register.subject` to create subjects and `@register.data` to create different data types.",
    requires = ''
)

import gc
import random
from .trans import Transformation
from .stdio import IMG
from .zxhtools.TRS import TRS
try: from PIL import Image; available_Image = True
except ImportError: available_Image = False

with __info__:
    import batorch as bt
    from pycamia import alias, Path, SPrint, ByteSize
    from pycamia import avouch, touch, to_tuple, arg_tuple, execblock

__all__ = """
    Key
    DataObject
    Slicer
    Dataset
""".split()
    # Subject
    # ImageObject
    # Dataset
    # MedicalDataset
    
def deapprox(x):
	if x.__class__.__name__ == 'approx_T': return x.__class__.__mro__[1](x)
	return x

def approx(x):
	if x == NotImplemented: return x
	class approx_T(x.__class__):
		for __op__ in [f"__{op}__" for op in """
				add radd iadd sub rsub isub mul rmul imul
				truediv rtruediv itruediv floordiv rfloordiv ifloordiv
				mod rmod imod pow rpow ipow lshift rshift
			""".split() if op]:
			exec(f"def {__op__}(self, other): return approx(getattr(deapprox(self), '{__op__}')(deapprox(other)))")
		def __str__(self): return '≈' + super().__str__()
	return approx_T(x)

class SortedDict(dict):
    
    def __new__(cls, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], dict):
            return SortedDict.__new__dict(*args, **kwargs)
        if len(args) > 1 and not isinstance(args[0], str):
            return SortedDict.__new__keyvalue(*args, **kwargs)
        return SortedDict.__new__default(*args, **kwargs)

    @classmethod
    def __new__keyvalue(cls, keys:list=[], values:list=[]):
        self = super().__new__(cls, zip(keys, values))
        self.key_order = keys
        return self

    @classmethod
    def __new__dict(cls, dic:dict, order=None):
        self = super().__new__(cls, dic)
        if order is None: self.key_order = list(dic.keys())
        else: self.key_order = order
        return self

    @classmethod
    def __new__default(cls, *args, **kwargs):
        self = super().__new__(cls, kwargs)
        if len(args) == 0: self.key_order = list(kwargs.keys())
        else: self.key_order = list(arg_tuple(args))
        return self

    def sort(self): self.key_order.sort()

    def shuffle(self): random.shuffle(self.key_order)
        
    def first(self): return self.key_order[0], self[self.key_order[0]]

    def keys(self):
        for k in self.key_order: yield k

    def values(self):
        for k in self.key_order: yield self[k]

    def items(self):
        for k in self.key_order: yield (k, self[k])

    def setdefault(self, k, v):
        if k not in self: self.key_order.append(k)
        super().setdefault(k, v)

    def pop(self, k, value):
        if k in self.key_order: self.key_order.remove(k)
        return super().pop(k, value)

    def __setitem__(self, k, v):
        if k not in self: self.key_order.append(k)
        super().__setitem__(k, v)

    def update(self, dic):
        for k, v in dic.items(): self[k] = v
        
    def copy(self):
        return SortedDict(super().copy(), order = self.key_order)
    
    def __iter__(self): return self.keys()
    
    @alias('__repr__')
    def __str__(self): return '{' + ', '.join([f"{repr(k)}: {repr(v)}" for k, v in self.items()]) + '}'
    
    def __len__(self): return len(self.key_order)
    
    def get_key(self, i): return self.key_order[i]

    def get_value(self, i): return self[self.key_order[i]]
    
    def get_item(self, i): k = self.key_order[i]; return (k, self[k])
    
    def __getitem__(self, k):
        if isinstance(k, slice):
            key_order = self.key_order[k]
            return SortedDict({i: self[i] for i in key_order}, key_order)
        return_value = super().get(k, None)
        if return_value is None and isinstance(k, int): return_value = self[self.get_key(k)]
        return return_value
    
def is_subkey(key_sub, key_cons):
    for k, v in key_cons.items():
        if v == ...: continue
        if isinstance(v, (list, tuple)):
            cv = key.get(k, None)
            if cv in v: ...
            elif cv == v: ...
            else: return False
        if key_sub.get(k, None) != v: return False
    return True
    
class Key:
    
    main_key_name = None
    
    @classmethod
    def __class_getitem__(cls, main_key_name):
        cls.main_key_name = main_key_name
        return cls
    
    def __init__(self, **kwargs):
        self._keys = []
        self._is_hidden = []
        self._hidden_values = {}
        self._visible_values = {}
        for k, v in kwargs.items():
            if k.startswith('_'):
                self._keys.append(k[1:])
                self._is_hidden.append(True)
                self._hidden_values[k[1:]] = v
            else:
                self._keys.append(k)
                self._is_hidden.append(False)
                self._visible_values[k] = v
                
    def update(self, **kwargs):
        for k, v in kwargs.items():
            is_hidden = False
            if k.startswith('_'): k = k[1:]; is_hidden = True
            elif k in self._keys: is_hidden = self._is_hidden[self._keys.index(k)]
            if is_hidden:
                self._hidden_values[k] = v
            else: self._visible_values[k] = v
            
    def pop(self, k):
        i = self._keys.index(k)
        self._keys.pop(i)
        is_hidden = self._is_hidden.pop(i)
        if is_hidden: return self._hidden_values.pop(k)
        return self._visible_values.pop(k)
            
    def tuple(self):
        return tuple(self[k] for k in self._keys)
    
    def __getitem__(self, k):
        if isinstance(k, int): k = self._keys[k]
        if k.startswith('_'): return self._hidden_values[k[1:]]
        if k in self._keys:
            if self._is_hidden[self._keys.index(k)]:
                return self._hidden_values[k]
            else: return self._visible_values[k]
        raise IndexError(f"No '{k}' in Key object. ")
    
    def __setitem__(self, k, v):
        if isinstance(k, int): k = self._keys[k]
        if k.startswith('_'):
            self._hidden_values[k[1:]] = v
            if k[1:] not in self._hidden_values:
                self._keys.append(k[1:])
            return
        if k in self._keys:
            if self._is_hidden[self._keys.index(k)]:
                self._hidden_values[k] = v
            else: self._visible_values[k] = v
        else:
            self._keys.append(k)
            self._visible_values[k] = v
            
    def __len__(self): return len(self._keys)
    
    def __contains__(self, k): return k.lstrip('_') in self._keys
    
    def keys(self): return ['_' if h else '' + k for h, k in zip(self._is_hidden, self._keys)]
    def visible_keys(self): return [k for h, k in zip(self._is_hidden, self._keys) if not h]
    def hidden_keys(self): return [k for h, k in zip(self._is_hidden, self._keys) if h]
    def values(self): return [self[k] for k in self._keys]
    def items(self): return zip(self.keys(), self.values())
    def get(self, k, v): return self[k] if k in self else v
    def dict(self): return dict(self.items())
    def copy(self): return Key(**self.dict())
    for __op__ in [f'__{op}__' for op in "lt le gt ge eq".split()]:
        execblock(f"""
        def {__op__}(x, y):
            return x.tuple().{__op__}(getattr(y, 'tuple', lambda: tuple(y))())
        """)

    def __hash__(self): return hash(self.tuple())
    
    @alias('__repr__')
    def __str__(self):
        main_key_list = [f"[{self.main_key_name}]: {repr(self.main_key)}"] if self.main_key_name in self._keys else []
        return '{' + ', '.join(main_key_list + [('({})' if h else '{}').format(f"{k}: {repr(self[k])}") for k, h in zip(self._keys, self._is_hidden) if k != self.main_key_name]) + '}'
    
    def __call__(self, **kwargs):
        self = self.copy()
        for k, v in kwargs.items():
            avouch(k in self._keys, f"cannot change '{k}' in Key as no such property is defined. ")
            self[k] = v
        return self
    
    @property
    def main_key(self):
        return self.get(self.main_key_name, None)

# class Info:
#     def __init__(self, **kwargs):
#         self._keys = []
#         self._values = []
#         self._hidden_keys = []
#         self._hidden_values = []
#         for k, v in kwargs.items():
#             if k.startswith('_'): 
#                 self._hidden_keys.append(k[1:])
#                 self._hidden_values.append(v)
#                 continue
#             self._keys.append(k)
#             self._values.append(v)

#     def update(self, **kwargs):
#         for k, v in kwargs.items():
#             if k in self._keys:
#                 self._values[self._keys.index(k)] = v
#                 continue
#             if k.startswith('_') and k[1:] in self._hidden_keys:
#                 self._hidden_values[self._hidden_keys.index(k[1:])] = v
#                 continue
#             if k.startswith('_'): 
#                 self._hidden_keys.append(k[1:])
#                 self._hidden_values.append(v)
#                 continue
#             self._keys.append(k)
#             self._values.append(v)

#     def tuple(self):
#         return tuple(self._values)
    
#     def visible(self):
#         return self._keys

#     def keys(self):
#         return self._keys + ['_'+k for k in self._hidden_keys]

#     def values(self):
#         return self._values + self._hidden_values
    
#     def items(self): return zip(self.keys(), self.values())
    
#     def get(self, k, v): return self[k] if k in self else v
    
#     def dict(self): return dict(self.items())
        
#     def copy(self): return Info(**self.dict())

#     def __getitem__(self, k):
#         if isinstance(k, int): return self.tuple()[k]
#         elif k in self._keys: return self._values[self._keys.index(k)]
#         elif k in self._hidden_keys: return self._hidden_values[self._hidden_keys.index(k)]
#         elif k.startswith('_') and k[1:] in self._hidden_keys: return self._hidden_values[self._hidden_keys.index(k[1:])]
#         else: raise IndexError(f"No '{k}' in Info object. ")

#     def __setitem__(self, k, v):
#         if k in self._keys: self._values[self._keys.index(k)] = v
#         elif k in self._hidden_keys: self._hidden_values[self._hidden_keys.index(k)] = v
#         elif k.startswith('_') and k[1:] in self._hidden_keys: self._hidden_values[self._hidden_keys.index(k[1:])] = v
#         elif k.startswith('_'): self._hidden_keys.append(k[1:]); self._hidden_values.append(v)
#         else: self._keys.append(k); self._values.append(v)

#     def __len__(self): return len(self._keys) + len(self._hidden_keys)

#     def __contains__(self, k): return k in self._keys or k in self._hidden_keys or k.startswith('_') and k[1:] in self._hidden_keys

#     for __op__ in [f'__{op}__' for op in "lt le gt ge eq".split()]:
#         execblock(f"""
#         def {__op__}(x, y):
#             return x.tuple().{__op__}(getattr(y, 'tuple', lambda: tuple(y))())
#         """)

#     def __hash__(self): return hash(self.tuple())
    
#     @alias('__repr__')
#     def __str__(self):
#         return '{' + ', '.join([f"{k}: {repr(v)}" for k, v in zip(self._keys, self._values)]) + \
#             ('; ' + ', '.join([f"{k}: {repr(v)}" for k, v in zip(self._hidden_keys, self._hidden_values)]) if len(self._hidden_keys) > 0 else '') + '}'

# class Subject:
#     def __init__(self, x=None, **kwargs):
#         if x is not None and isinstance(x, (dict, Info)):
#             kwargs = x
#         self.subject = Info(**kwargs)
#         self.image = Info(_n_subimage = 1)
#         self.subimage = Info(subimage_id = 'whole')

#     def Image(self, **kwargs):
#         if 'n_subimage' in kwargs: kwargs['_n_subimage'] = kwargs.pop('n_subimage')
#         if 'path' in kwargs: kwargs['_path'] = kwargs.pop('path')
#         self.image.update(**kwargs)
#         return self

#     def SubImage(self, *args, **kwargs):
#         if len(args) == 1:
#             self.Image(n_subimage = args[0])
#             return self
#         elif len(args) == 0:
#             self.subimage.update(**kwargs)
#             return self
#         else: raise TypeError()

#     def update(self, **kwargs):
#         for k, v in kwargs.items():
#             if k in self.subject: self.subject[k] = v
#             elif k in self.image: self.image[k] = v
#             elif k in self.subimage: self.subimage[k] = v
#         return self

#     def get_subimage(self, i):
#         return Info(subimage_id = f"%0{len(str(self.image['n_subimage']))}d"%i)

#     def to_subimage(self, i):
#         ret = self.copy()
#         if i < 0: ret.update(subimage_id = 'whole')
#         else: ret.update(subimage_id = f"%0{len(str(self.image['n_subimage']))}d"%i)
#         return ret
    
#     def to_wholeimage(self):
#         return self.to_subimage(-1)

#     def tuple(self):
#         return self.subject.tuple() + self.image.tuple() + self.subimage.tuple()

#     def keys(self):
#         return self.subject.keys() + self.image.keys() + self.subimage.keys()
        
#     def copy(self):
#         subject = Subject(**self.subject.dict())
#         subject.image = self.image.copy()
#         subject.subimage = self.subimage.copy()
#         return subject

#     def __contains__(self, k):
#         if isinstance(k, int): return k < len(self)
#         elif k in self.subject: return True
#         elif k in self.image: return True
#         elif k in self.subimage: return True
#         else: return False

#     def __getitem__(self, k):
#         if isinstance(k, int): return self.tuple()[k]
#         elif k in self.subject: return self.subject[k]
#         elif k in self.image: return self.image[k]
#         elif k in self.subimage: return self.subimage[k]
#         else: raise IndexError(f"No '{k}' in micomputing.Subject-Image-SubImage-Info object. ")

#     def __setitem__(self, k, v):
#         if k in self.subject: self.subject[k] = v
#         elif k in self.image: self.image[k] = v
#         elif k in self.subimage: self.subimage[k] = v
#         else: raise IndexError(f"No '{k}' in micomputing.Subject-Image-SubImage-Info object. ")

#     def __len__(self): return len(self.subject) + len(self.image) + len(self.subimage)

#     for __op__ in [f'__{op}__' for op in "lt le gt ge eq".split()]:
#         execblock(f"""
#         def {__op__}(x, y):
#             return x.tuple().{__op__}(getattr(y, 'tuple', lambda: tuple(y))())
#         """)

#     def __hash__(self): return hash(self.tuple())
    
#     @alias('__repr__')
#     def __str__(self): return '{' + ' | '.join([str(self.subject).strip('{}'), str(self.image).strip('{}'), str(self.subimage).strip('{}')]) + '}'
    
#     def __call__(self, **kwargs):
#         subject = self.copy()
#         for k, v in kwargs.items(): subject[k] = v
#         return subject
    
class DataObject:
    
    preset_loaders = [IMG, TRS.load, Transformation.load, lambda x: x.open().read(), lambda x: open(x).read()] + [Image.open] if available_Image else []
    
    def __init__(self, path, loader=None, *, i_slice=None):
        self.path = Path(path)
        self.i_slice = i_slice
        self.loader = loader
        self._data = None
        
    @property
    def is_loaded(self): return self._data is not None

    def load(self, loader=None):
        if self.is_loaded: return self._data
        path = self.path.abs
        if loader is None and self.loader is not None: loader = self.loader
        if loader is None:
            for loader in DataObject.preset_loaders:
                try: self._data = loader(path); break
                except Exception as e:
                    if isinstance(e, (TypeError, ValueError, AssertionError)): pass
                    elif isinstance(e, FileNotFoundError): raise e
                    elif 'DecodeError' in str(type(e)): pass
                    else: raise e
            self.final_loader = ["IMG", "TRS", "Trans", "Path", "txt", "Image"][DataObject.preset_loaders.index(loader)]
            if self._data is None: raise TypeError(f"Cannot open file {path} yet (only *.nii, *.nii.gz, *.dcm, *.ima, *.nrrd, *.jpg, *.png, *.txt, *.trs, *.AFF, *.FFD, *.DDF are allowed), please contact the developpers (Error Code: D651). ")
        else:
            self._data = loader(path)
            self.final_loader = "DIY"
            return self._data
        if self.i_slice is not None:
            avouch(loader == IMG, "Only image objects can be breaked into slices. ")
            orient = self._data.orientation
            for a in 'IS':
                if a in orient: new_orient = orient.replace(a, '') + a
            self.affine = bt.tensor(self._data.affine)
            self.header = self._data.header
            self._data = self._data.reorient(new_orient).to_tensor()
            self.n_slice = self._data.size(-1)
            self._data = self._data[..., self.i_slice]
        elif loader == IMG:
            self.affine = bt.tensor(self._data.affine)
            self.header = self._data.header
            self._data = self._data.to_tensor()
            self.n_slice = self._data.size(-1)
        return self._data
    
    def slice(self, i_slice):
        self.i_slice = i_slice
        return self
    
    def __getitem__(self, i_slice=None):
        if i_slice is None: i_slice = self.i_slice
        else: avouch(self.i_slice is None, "Only unsliced IMG DataObject can be sliced by subscript. ")
        if not self.is_loaded:
            self.load(loader=IMG)
            orient = self._data.orientation
            for a in 'IS':
                if a in orient: new_orient = orient.replace(a, '') + a
            self.affine = bt.tensor(self._data.affine)
            self.header = self._data.header
            self._data = self._data.reorient(new_orient).to_tensor()
            self.n_slice = self._data.size(-1)
        return self._data[..., i_slice]

    def release(self):
        del self._data
        gc.collect()
        self._data = None

    @alias('__repr__')
    def __str__(self):
        if self.is_loaded: load_state = f"loaded by {self.final_loader}"
        else: load_state = "not loaded"
        slice_str = ''
        if self.i_slice is not None: slice_str = f" (slice {i_slice})"
        return f"<DataObject at {self.path.filename}{slice_str} [{load_state}]>"
    
    @property
    def data(self):
        if self._data is None: self.load()
        return self.__getitem__() if self.i_slice is not None else self._data
    
class Slicer:
    
    def __init__(self, path, loader=None):
        self.data_obj = DataObject(path, loader=loader)
        self.valid_slices = None
        self.slice_pointer = 0
        
    def items(self, key=Key(slice_id=None)):
        i_slice = 0
        while True:
            yield key(slice_id=i_slice), self.data_obj[i_slice]
            i_slice += 1
            if self.is_loaded and i_slice >= self.n_slice: break
        self.release()
        
    @property
    def is_loaded(self): return self.data_obj.is_loaded
    def load(self, loader=None):
        self.slice_pointer = 0
        return self.data_obj.load(loader=loader)
    def release(self): return self.data_obj.release()
    
    def __getitem__(self, index):
        if not self.data_obj.is_loaded: self.slice_pointer = 0
        if self.valid_slices is not None:
            if callable(self.valid_slices):
                self.valid_slices = self.valid_slices(self.data_obj.data)
            avouch(isinstance(self.valid_slices, list))
            index = self.valid_slices[index]
        return self.data_obj[index]
    
    def slice(self, i_slice):
        return self.data_obj.slice(i_slice)
    
    @property
    def n_slice(self):
        if self.valid_slices is None: return getattr(self.data_obj, 'n_slice', None)
        return len(self.valid_slices) if isinstance(self.valid_slices, list) else None

    @alias('__repr__')
    def __str__(self):
        if self.is_loaded: load_state = f"loaded by {self.final_loader}"
        else: load_state = "not loaded"
        return f"<SlicerObject at {self.path.filename} [{load_state}]>"
    
    def __getattribute__(self, name):
        if name in ("path", "loader", "final_loader", "i_slice", 
                    "is_loaded", "header", "affine", "data"):
            return getattr(self.data_obj, name)
        return super().__getattribute__(name)
    
    def set_valid_slices(self, obj):
        """
        Set the slices that are valid to pick from. 

        Args:
            obj (str or list or func): object determining the valid slice indices. 
                str: path to a file with content of format "s1, s2, ..., sn"
                list: a list of slice indices
                func: a function that returns a list of valid slice indices during data loading
                None: all slices are valid
        """
        if isinstance(obj, str):
            obj = Path(obj)
            avouch(obj.exists(), FileNotFoundError(f"Unknown file path '{obj}'. "))
            with open(obj) as fp:
                content = fp.read().strip()
                if content: self.valid_slices = list(eval(content))
                else: content = []
        elif isinstance(obj, list):
            self.valid_slices = obj
        elif callable(obj): self.valid_slices = obj
        else: raise TypeError(f"Unknown valid slice type {obj}. ")
        return self

class Dataset:
    
    def __init__(self, *args, name = None, main_key_name = 'subject_id', batch_pattern = 'unpaired'):
        """
        Create a dataset object. 
        
        Args:
            paths (str): The main folders of the dataset. Dataset will browse all the files within. 
            name (str): The name of the dataset, one can also identify it in the function name of the loading function. 
            main_key_name (str): The main key defined to distinguish different subjects. Defaults to 'subject_id'. 
            batch_pattern (str: unpaired/ paired/ random or list[Key]): Whether the images inside are paired. Defaults to 'unpaired'.
                unpaired: Values with different Key should be taken as independent data. 
                paired: Only values with different 'main_key' should be regarded as independent data, the input for batch function would be a tuple of such data. 
                random: Values with different Key are regarded as independent data, but the input for batch function would randomly pick the data of the same non-main_key keys. 
                list[Key]: A list of Key objects to constraint each element of the batch input. 
                    unpaired <==> Key()
                    paired   <==> e.g. [Key(subject_id=..., modality='CT', is_label=False), Key(subject_id=..., modality='T1', is_label=False), Key(subject_id=..., modality='T1', is_label=True)] # ... here is Ellipsis indicating the same subject_id
                    random   <==> e.g. [Key(modality='CT', is_label=False), Key(modality='T1', is_label=False), Key(modality='T1', is_label=True)]
            Note: the training/ validation/ test sub-datasets are split according to the 'main_key'.
        
        Examples::
            >>> from micomputing import data
            >>> @data.Dataset("folderpath1", "folderpath2")
            ... def <datasetname>(path):
            ...     '''
            ...     The main function maps path to a Key-value pair
            ...     Note that:
            ...     1) A list should be returned if multiple elements are obtained from a single path, e.g.,
            ...            return [Key(patient_id = path.split()[1], modality = path.name, slice_id = i), DataObject(path, i_slice=i) for i in range(n_slice)]
            ...     2) A single pair of Key-value is also allowed. 
            ...          The value should be replaced by a Slicer object for image slices, which can generate slice data after the image is loaded.
            ...          In which case the Key object should leave out namescope 'slice_id' by a default value None to be filled afterwards, e.g., 
            ...            return Key(patient_id = path.split()[1], modality = path.name, slice_id = None), Slicer(path)
            ...     3) No identical keys are allowed. One needs to set another property even for elements from a single datum, e.g.,
            ...            Key(..., slice_id = xx), DataObject(path) for '*.dcm' or '*.ima' slice files. 
            ...     4) None should be returned if the path is to be omitted. 
            ...     '''
            ...     return Key(patientID = path.split()[1], modality = path.name), DataObject(path)
            ... 
            >>> <datasetname>
            <datasetname> Dataset (121 images): 
            ==================================================
            patientID = 152
            || modality = MR
            || modality = CT
            ...
            patientID = 174
            || modality = CT
        """
        self.n_slice_record = []
        self.use_slice_elements = False
        self.name = name
        self.main_key_name = main_key_name
        self.batch_pattern = batch_pattern
        self.subject_index = {}
        self.batch_index = {}
        self._training_set = []
        self._validation_set = []
        self._testing_set = []
        if len(args) == 1:
            x = args[0]
            if isinstance(x, dict): self.__init__dict(*args)
            elif isinstance(x, str): self.__init__str(*args)
            else: self.__init__sequence(*args)
        else: self.__init__str(*args)

    def __init__dict(self, x: SortedDict):
        self.data = x
        self.batch_pointer = {'training': {}, 'validation': {}, 'testing': {}}
        self.directories = []
        for k in self.data.keys():
            if 'path' not in k: continue
            if k.path.ref not in self.directories: self.directories.append(k.path.ref)
        self._cache = []

    def __init__sequence(self, x: (list, tuple)):
        self.__init__(*x)

    def __init__str(self, *directories: str):
        self.data = SortedDict()
        self.batch_pointer = {'training': {}, 'validation': {}, 'testing': {}}
        self.directories = map(Path, directories)
        self._cache = []

    def __call__(self, func):
        self.path2key_value = func
        if self.name is None:
            self.name = func.__name__
        self.sort_files()
        return self

    def __len__(self):
        self.check_data()
        return len(self.data)

    @alias("manual_seed")
    def seed(self, s): random.seed(s)

    def check_data(self): avouch(len(self.data) > 0, "Dataset not created yet. Use `@Dataset(directory_paths)` in front of a function mapping a path to key structure to create Dataset. ")

    def cache(self, k, v=None):
        if v is None: return self._cache[k]
        self._cache[k] = v

    def byte_size(self):
        total_size = 0
        for ele in self.data.values():
            if isinstance(ele, Dataset):
                total_size += ele.byte_size()
            elif isinstance(ele, DataObject):
                if isinstance(ele.data, bt.Tensor):
                    total_size += ele.data.byte_size()
            elif isinstance(ele, bt.Tensor):
                total_size += ele.byte_size()
        return ByteSize(total_size)
    
    def append_n_slice_record(self, new_sample):
        n_sample = 20
        if len(self.n_slice_record) < n_sample: self.n_slice_record.append(new_sample); return
        i = [r < new_sample for r in self.n_slice_record].index(False)
        self.n_slice_record.insert(i, new_sample)
        if i < n_sample // 2: self.n_slice_record.pop(0)
        else: self.n_slice_record.pop(-1)
    
    @property
    def estimated_n_slice(self):
        if len(self.n_slice_record) > 0: return approx(round(sum(self.n_slice_record) / len(self.n_slice_record)))
        else: return approx(20)

    def sort_files(self):
        for d in self.directories:
            for f in d.iter_files():
                ret = self.path2key_value(f)
                if ret is None: continue
                if isinstance(ret, tuple):
                    key, value = ret
                    if key.main_key_name is None: key.__class__.main_key_name = self.main_key_name
                    if isinstance(value, Slicer): key['slice_id'] = None; self.use_slice_elements = True
                    self.data[key] = value
                    self.subject_index.setdefault(key.main_key, [])
                    self.subject_index[key.main_key].append(key)
                elif isinstance(ret, list):
                    for key, value in ret:
                        if key.main_key_name is None: key.__class__.main_key_name = self.main_key_name
                        self.data[key] = value
                        self.subject_index.setdefault(key.main_key, [])
                        self.subject_index[key.main_key].append(key)
                else: raise TypeError(f"The 'path2key_value' method decorated by @Dataset(...) should return Key-value pair(s), but got {ret}. ")
        if self.batch_pattern == 'unpaired': self.batch_pattern = Key()
        elif isinstance(self.batch_pattern, str):
            all_items = []
            for k in self.data:
                if self.batch_pattern == 'paired': k = k(**{k.main_key_name: ...})
                else: k = k.pop(k.main_key_name)
                if k not in all_items: all_items.append(k)
            self.batch_pattern = all_items
        if isinstance(self.batch_pattern, Key): self.batch_pattern = [self.batch_pattern]
        avouch(isinstance(self.batch_pattern, list), f"Property batch_pattern={self.batch_pattern} is invalid for Dataset object. ")
        
        remove_list = []
        for bk in self.batch_pattern: # for each batch constraint
            self.batch_index.setdefault(bk, SortedDict())
            for sid in self.main_keys:
                if sid in remove_list: continue
                paired = bk.get(bk.main_key_name, None) == ...
                if paired: # if the element needs paired
                    candidates = [k for k in self.subject_index[sid] if is_subkey(k, bk(**{bk.main_key_name: sid}))]
                else: # no paired element needed
                    candidates = [k for k in self.subject_index[sid] if is_subkey(k, bk)]
                if len(candidates) > 0: self.batch_index[bk][sid] = candidates; continue # subimage found
                elif paired: remove_list.append(sid) # paired images missing member
        if len(remove_list) == self.n_main_key or any(len(cand) == 0 for cand in self.batch_index.values()):
            raise TypeError(f"Cannot find required image output following batch pattern {self.batch_pattern} for {self.name} dataset. ")
        for sid in remove_list:
            for k in self.subject_index[sid]: self.data.pop(k, None) # remove from dataset
            for bk in self.batch_pattern: self.batch_index[bk].pop(sid, None) # remove from content index

        self.data.sort()
        self.split_datasets(training=0.7, validation=0.2)
    
    @property
    def main_keys(self):
        return list(set(k.main_key for k in self.data.keys()))
    
    @alias("n_subject")
    @property
    def n_main_key(self): return len(self.main_keys)

    def __str__(self) -> str:
        self.check_data()
        self.data.sort()
        str_print = SPrint()
        s = lambda n: "s" if n > 1 else ''
        item = 'slice' if self.use_slice_elements else 'image'
        str_print(f"{'' if self.name is None else self.name} Dataset", f"({self.n_data} {item}{s(self.n_data)}, {self.n_subject} subject{s(self.n_subject)}): ")
        str_print('=' * 50)
        if self._training_set:
            str_print(
                f"{self.n_training} training {item}{s(self.n_training)} ({self.n_training_subject} subject{s(self.n_training_subject)})", 
                f"{self.n_validation} validation {item}{s(self.n_validation)} ({self.n_validation_subject} subject{s(self.n_validation_subject)})", 
                f"{self.n_testing} test {item}{s(self.n_testing)} ({self.n_testing_subject} subject{s(self.n_testing_subject)}).", sep=', ')
            str_print('=' * 50)
        
        n_main_keys = len(set([k.main_key for k in self.data.keys()]))
        should_omit = self.n_main_key > 6 # whether to omit the middle data. 
        count = 0 # the counter of number of data, used for omission. 
        prefix = lambda i: ' |  ' * i
        prev_key = None
        for d, v in self.data.items():
            first_diff = None
            for i, k in enumerate(d.visible_keys()):
                if first_diff is not None and i > first_diff or prev_key is None or prev_key[k] != d[k]:
                    if first_diff is None: first_diff = i
            prev_key = d
            if should_omit and first_diff == 0:
                count += 1
                if count == 3: str_print('...')
                if count > self.n_main_key - 2: count = 0
            if should_omit and count > 2: continue
            for i, k in enumerate(d.visible_keys()):
                if i < first_diff: continue
                if i < len(d) - 1: str_print(prefix(i), f"{k} = {d[k]}", sep=''); continue
                skey = f"{k} = {d[k]}: "
                str_v = str(v).split('=' * 10)[-1].lstrip('=').strip()
                str_print(prefix(i), skey, str_v.replace('\n', '\n' + prefix(i) + ' ' * len(skey)), sep='')
        return str_print.text

    def setdefault(self, k, v):
        if k not in self.data: self.data[k] = v

    def get_key(self, *args, **kwargs):
        avouch(len(args) > 0 and len(kwargs) == 0 or len(args) == 0 and len(kwargs) > 0)
        if len(args) > 0:
            args = to_tuple(args)
            if len(args) == 1 and isinstance(args[0], (list, tuple)): args = args[0]
        candidates = []
        for k in self.data.keys():
            if len(args) > 0:
                if all([x in k.tuple() for x in args]):
                    candidates.append(k)
            else:
                if all([k.get(a, None) == kwargs[a] for a in kwargs]):
                    candidates.append(k)
        if len(candidates) == 1: return candidates[0]
        raise TypeError(f"Cannot use incomplete of keys {args if len(args) > 0 else kwargs} to find non-unique or non-existed items. Number of matches: {len(candidates)} ")
    
    def get(self, *args, **kwargs):
        """
        get(key, default_value) OR get(constraint_key_name=constraint_value)
        """
        avouch(len(args) > 0 and len(kwargs) == 0 or len(args) == 0 and len(kwargs) > 0)
        if len(args) > 0:
            k = args[0]
            v = None
            if len(args) > 1: v = args[1]
            if k not in self.data: return v
            return self.data[k]
        try: return self.data[self.get_key(**kwargs)]
        except TypeError: raise KeyError(f"Invalid key {kwargs}.")

    def items(self): return self.data.items()

    def first(self): return self.data.first()

    def shuffle(self): return self.data.shuffle()

    def __setitem__(self, k, v):
        self.data[k] = v

    def __getitem__(self, k):
        if isinstance(k, Key) and k == Key(): return self
        if isinstance(k, list): return Dataset(
            SortedDict({l: self[l] for l in k}), 
            name=self.name+'[subset]', 
            main_key_name=self.main_key_name, 
            batch_pattern=self.batch_pattern
        )
        if k in self.data: return self.data[k]
        if isinstance(k, (dict, Key)): return self.data[self.get_key(**k)]
        try: return self.data[self.get_key(k)]
        except TypeError: raise KeyError(f"Invalid key {k}.")
    
    def __getattr__(self, k):
        try: return getattr(super(), k, None)
        except KeyError as e: raise AttributeError(str(e))

    def __iter__(self):
        return iter(self.data)

    def split_datasets(self, training = None, validation = None, testing = None):
        """
        Split dataset to training, validation and test sets by ratio. e.g. split_dataset(training=0.8, validation = 0.1)
        """
        self.check_data()
        if validation is None and (training is None or testing is None): validation = 0.
        if testing is None and training is None: testing = 0.
        if training is None: training = 1. - validation - testing
        elif validation is None: validation = 1. - testing - training
        elif testing is None: testing = 1. - training - validation
        avouch(abs(training + testing + validation - 1.) < 1e-3, "Invalid ratios for function 'split_datasets' (Sum is not 1). ")
        main_keys = self.main_keys
        n_train = int(training * len(main_keys))
        n_valid = int(validation * len(main_keys))
        random.shuffle(main_keys)
        def add_subimage(main_key_set):
            return sum((self.subject_index[mk] for mk in main_key_set), [])
        self._training_set = add_subimage(main_keys[:n_train])
        self._validation_set = add_subimage(main_keys[n_train:n_train + n_valid])
        self._testing_set = add_subimage(main_keys[n_train + n_valid:])
        return self
    
    def needed_in_batch(self, key, batch_pattern=None):
        if batch_pattern is None: batch_pattern = self.batch_pattern
        for const in batch_pattern:
            if is_subkey(key, const): return True
        return False
    
    @property
    def n_training_subject(self): return len(list(set(k.main_key for k in self._training_set))) if self._training_set else None
    @property
    def n_validation_subject(self): return len(list(set(k.main_key for k in self._validation_set))) if self._validation_set else None
    @property
    def n_testing_subject(self): return len(list(set(k.main_key for k in self._testing_set))) if self._testing_set else None
    
    @property
    def n_data(self):
        if self.data is None: return
        total_item = 0
        for key in self.data:
            if not self.needed_in_batch(key): continue
            if self.use_slice_elements:
                if not self.data[key].is_loaded: total_item = total_item + self.estimated_n_slice
                else: total_item += self.data[key].n_slice
            else: total_item += 1
        return total_item
    @property
    def n_training(self):
        if not self._training_set: return
        total_item = 0
        for key in self.data:
            if key not in self._training_set: continue
            if not self.needed_in_batch(key): continue
            if self.use_slice_elements: 
                if not self.data[key].is_loaded: total_item = total_item + self.estimated_n_slice
                else: total_item += self.data[key].n_slice
            else: total_item += 1
        return total_item
    @property
    def n_validation(self):
        if not self._validation_set: return
        total_item = 0
        for key in self.data:
            if key not in self._validation_set: continue
            if not self.needed_in_batch(key): continue
            if self.use_slice_elements: 
                if not self.data[key].is_loaded: total_item = total_item + self.estimated_n_slice
                else: total_item += self.data[key].n_slice
            else: total_item += 1
        return total_item
    @property
    def n_testing(self):
        if not self._testing_set: return
        total_item = 0
        for key in self.data:
            if key not in self._testing_set: continue
            if not self.needed_in_batch(key): continue
            if self.use_slice_elements: 
                if not self.data[key].is_loaded: total_item = total_item + self.estimated_n_slice
                else: total_item += self.data[key].n_slice
            else: total_item += 1
        return total_item

    @alias("train_batch")
    def training_batch(self, n, **kwargs): return self.batch('training', n, **kwargs)
    @alias("valid_batch")
    def validation_batch(self, n, **kwargs): return self.batch('validation', n, **kwargs)
    @alias("test_batch")
    def testing_batch(self, n, **kwargs): return self.batch('testing', n, **kwargs)
    
    def batch(self, stage='training', n_batch=4, shuffle=True, drop_last=True, none_each_epoch=True, restart=False):
        """
        Create a batch data. 

        Args:
            stage (str, 'training'|'validation'|'testing'): The stage we are in, determining the subset we are using. Defaults to 'training'.
            n_batch (int): The size of batch. Defaults to 4.
            shuffle (bool): Whether to randomly pick data from the dataset or not. Defaults to True.
                [True]: The dataset will automatically shuffle for each epoch. 
                [False]: No randomization is used in selecting batches, they are selected in order. 
            drop_last (bool):
                Whether to drop the last data which is less than a batch. Defaults to True.
                [True]: Drop the remaining data.
                [False]: Use the remaining data to create a batch with a batch size smaller than normal ones. 
            none_each_epoch (bool): Whether to output a None value at the end of each epoch. Defaults to True.
            restart (bool): Manually use a restart of epoch for this batch. Defaults to False.
        """
        self.check_data()
        if not self._training_set: self.split_datasets(training=0.7, validation=0.2)
        stage = stage.lower()
        if stage == 'training': subset = self._training_set
        elif stage == 'validation': subset = self._validation_set
        elif stage == 'testing': subset = self._testing_set
        
        selected_items = []
        batch_list = []
        new_epoch = False
        release_list = []
        for bp in self.batch_pattern:
            candidates = self.batch_index[bp]
            cur_pointer = 0 if restart else self.batch_pointer[stage].get(bp, 0)
            if cur_pointer == 0 and self.use_slice_elements: cur_pointer = 0, 0
            batch_sids = []
            if bp.main_key == ... and selected_items:
                if self.use_slice_elements:
                    batch_list.append([self.data[random.choice(candidates[img.main_key])].slice(slc) for img, slc in selected_items])
                else: batch_list.append([self.data[random.choice(candidates[img.main_key])] for img in selected_items])
            else:
                cum_n_batch = 0
                if self.use_slice_elements:
                    cur_pointer, slice_pointer = cur_pointer
                    batched_slices = []
                    for i in range(n_batch):
                        image_key = random.choice(candidates[cur_pointer])
                        sample_image = self.data[image_key]
                        sample_image.load()
                        self.append_n_slice_record(sample_image.n_slice)
                        n_remain_batch = n_batch - len(batched_slices)
                        new_slice_pointer = slice_pointer + n_remain_batch
                        selected_items.extend([(image_key, i) for i in range(slice_pointer, min(new_slice_pointer, sample_image.n_slice))])
                        batched_slices.extend([sample_image.slice(i) for i in range(slice_pointer, min(new_slice_pointer, sample_image.n_slice))])
                        if new_slice_pointer >= sample_image.n_slice: new_slice_pointer = 0; cur_pointer += 1; release_list.append(sample_image)
                        if len(batched_slices) >= n_batch: break
                    batch_list.append(batched_slices)
                    if cur_pointer >= len(candidates): cur_pointer = 0; new_slice_pointer = 0; new_epoch = True
                    self.batch_pointer[stage][bp] = cur_pointer, new_slice_pointer
                else: 
                    new_pointer = cur_pointer + n_batch
                    selected_items = random.choice(candidates[cur_pointer: new_pointer])
                    batch_list.append([self.data[cand] for cand in selected_items])
                    if new_pointer >= len(candidates): new_pointer = 0; new_epoch = True
                    self.batch_pointer[stage][bp] = new_pointer

        is_partial_batch = any(len(x) < n_batch for x in batch_list)
        if is_partial_batch: new_epoch = True
        if new_epoch: 
            self.batch_pointer = {'training': {}, 'validation': {}, 'testing': {}}
            if shuffle:
                for bp in self.batch_pattern: self.batch_index[bp].shuffle()

        if is_partial_batch and drop_last:
            for r in release_list: r.release()
            if none_each_epoch: return
            else: return self.batch(stage=stage, n_batch=n_batch, shuffle=shuffle, 
                                    drop_last=drop_last, none_each_epoch=none_each_epoch, restart=restart)
        return_value = self.get_batch(batch_list)
        for r in release_list: r.release()
        return return_value
    
    def get_batch(self, batch_list):
        """
        The function that creates tuple of batch by structured batch data. 

        Args:
            batch_list (list): a list of length equal to *.batch_pattern; with each element a list of length n_batch. 
                The nested elements are DataObjects (Slicer output in major function would result in DataObjects with i_slice). 
        """
        res_list = []
        for output_batch in batch_list:
            ref = output_batch[0]
            if isinstance(ref, DataObject):
                output_batch = [x.data for x in output_batch]
                ref = ref.data
            elif isinstance(ref, bt.Tensor): ...
            else: res_list.append(output_batch); continue
            avouch(isinstance(ref, bt.Tensor))
            if ref.has_batch: res_list.append(bt.cat(output_batch, {}, crop=True).detach())
            else: res_list.append(bt.stack(output_batch, {}, crop=True).detach())
        return tuple(res_list)

# class MedicalSubImage(Dataset):
    
#     def __getitem__(self, k):
#         try:
#             if isinstance(k, Subject): return self.data[k.subimage]
#         except KeyError: raise KeyError(f"Invalid key {k}.")
#         return super().__getitem__(k)

#     def __setitem__(self, k, v):
#         if isinstance(k, Subject):
#             self.data.setdefault(k.subimage, v)
#         else: self.data[k] = v

#     @property
#     def IMG(self):
#         if len(self.data) == 1:
#             return self.data[next(self.data.keys())].load()
    
# class MedicalImage(Dataset):
    
#     def __getitem__(self, k):
#         try:
#             if isinstance(k, Subject): return self.data[k.image][k.subimage]
#         except KeyError as e: raise KeyError(eval(str(e)).strip('.') + f" in {k}.")
#         return super().__getitem__(k)

#     def __setitem__(self, k, v):
#         if isinstance(k, Subject):
#             self.data.setdefault(k.image, MedicalSubImage())
#             self.data[k.image].setdefault(k.subimage, v)
#         else: self.data[k] = v

#     @property
#     def IMG(self):
#         if len(self.data) == 1:
#             return self.data[next(self.data.keys())].load()

# class MedicalDataset(Dataset):

#     def __init__(self, *args, **kwargs):
#         """
#         register data type. 
        
#         Args:
#             paths (str): The folders with data. Dataset will browse all the files within and 
#                 use the decorated function to change a path into a Subject/ImageObject structure. 
#                 The arguments for Subject/ImageObject are their 'info'.
#             paired_subimage (bool): whether the images inside are paired, the 'paired' images 
#                 will not split in traing/validation/test datasets. Defaults to True.
        
#         Examples::
#             >>> from micomputing.data import MedicalDataset, Subject, Image
#             >>> @MedicalDataset("folderpath1", "folderpath2")
#             ... def datasetname(path):
#             ...     return Subject(patientID = path.split()[1]) \
#             ...           .Image(modality = path.name)
#             ... 
#             >>> datasetname
#             datasetname Dataset (121 images): 
#             ==================================================
#             patientID = 152
#             || modality = MR
#             || modality = CT
#             ...
#             patientID = 174
#             || modality = CT
#         """
#         super().__init__(*args, **kwargs)
#         self.paired_subimage = kwargs.get('paired_subimage', True)
#         self.template = None

#     def __str__(self) -> str:
#         self.check_data()
#         str_print = SPrint()
#         s = "s" if len(self.data) > 1 else ''
#         str_print(f"{self.name} Dataset ({len(self.data)} subject{s}): ")
#         str_print('=' * 50)
#         omit = len(self.data) > 8
#         start = lambda i: ' |  ' * i
#         for i, (sb, sbd) in enumerate(self.data.items()):
#             if not omit or i < 2 or i >= len(self.data) - 2:
#                 str_print('\n'.join([f"{k} = {v}" for k, v in sb.items()]) + ':')
#                 for im, imd in sbd.data.items():
#                     str_print('\n'.join([f"{start(1)}{k} = {v}" for k, v in im.items()]) + ':')
#                     if not isinstance(imd, Dataset):
#                         str_print(start(2), imd)
#                         continue
#                     for si, sid in imd.data.items():
#                         str_print('\n'.join([f"{start(2)}{k} = {v}" for k, v in si.items()]) + ':')
#                         str_print(start(3), sid)
#             if omit and i == 2: str_print('...')
#         return str_print.text

#     def __getitem__(self, k):
#         try:
#             if isinstance(k, Subject): return self.data[k.subject][k.image][k.subimage]
#         except KeyError: raise KeyError(f"Invalid key {k}.")
#         if k in self.data: return self.data[k]
#         if isinstance(k, (dict, Info)): return self.data[self.get_info(**k)]
#         return self.data[self.get_info(k)]

#     def __setitem__(self, k, v):
#         if isinstance(k, Subject):
#             self.data.setdefault(k.subject, MedicalImage())
#             self[k.subject].setdefault(k.image, MedicalSubImage())
#             self[k.subject][k.image].setdefault(k.subimage, v)
#         else: self.data[k] = v
    
#     def sort_files(self):
#         for d in self.directories:
#             f = next(d.iter_files())
#             if f | 'dcm' or f | 'ima': search_list = d.iter_subdirs()
#             else: search_list = d.iter_files()
#             for f in search_list:
#                 info = self.map_info(f)
#                 if info is None: continue
#                 info.Image(path=f)
#                 if self.template is None: self.template = info
#                 else:
#                     avouch(self.template.subject.keys() == info.subject.keys() and self.template.image.keys() == info.image.keys())
#                     if len(self.template.keys()) < len(info.keys()): self.template = info
#                     info.SubImage(**{k: '' for k in self.template.keys()[len(info.keys()):]})
#                 self[info] = ImageObject(f)
#         self.data.sort()
#         self.split_datasets(training=0.7, validation=0.2)

#     def select(self, func=None, **kwargs):
#         """
#         Select elements in the data. 
        
#         Note: One can use decorator `@datasetname.select` of a select function to perform an in-place select or 
#             use function `datasetname.select(some_property_need_filtering=filter_value)` to create a new Dataset.
            
#         Args:
#             func (callable): A filter that accepts `subject_info` and `data_for_subject`.
#                 subject_info (dict): The dict consist of all info for subject.
#                 data_for_subject (dict): The 'info' of all child images, as key to find them, as well as 
#                     information needed for filtering.

#         Examples::
#             >>> @datasetname.select
#             >>> def patientID(subject_info, data_for_subject):
#             ...     \"\"\"data_for_subject: SortedDict containing info: data objects. \"\"\"
#             ...     if subject_info['patientID'] == "72": return False # bad data
#             ...     all_modalities = [i['modality'] for i in data_for_subject]
#             ...     return 'CT' in all_modalities and 'MR' in all_modalities
#             ... 
#             >>> datasetname.select(modality='CT&MR')
#             datasetname Dataset (111 images): 
#             ==================================================
#             patientID = 152
#             || modality = MR
#             || modality = CT
#             ...
#             patientID = 162
#             || modality = MR
#             || modality = CT
#         """
#         self.check_data()
#         inplace = True
#         if func is None:
#             def selected(info, data):
#                 def sat(x, y):
#                     get_y = lambda u, v: touch(lambda: type(u)(v), v)
#                     if not isinstance(y, str): return any([i == get_y(i, y) for i in x])
#                     return any([(all([any([i == get_y(i, d) for i in x]) for d in c.split('&')]), print(c))[0] for c in y.split('|')])
#                 all_info = [info.copy().Image(**k.dict()) for k in data.keys()]
#                 if len(all_info) == 0:
#                     return all([sat([info[k]], v) for k, v in kwargs.items() if k in info])
#                 return all([sat([i[k] for i in all_info], v) for k, v in kwargs.items() if k in all_info[0]])
#             func = selected
#             inplace = False
#         to_delete = []
#         for info in map(Subject, self.data.keys()):
#             if not func(info, self.data[info.subject].data): to_delete.append(info.subject)
#         if inplace:
#             for i in to_delete[::-1]: self.data.pop(i)
#             self.data.sort()
#             self.split_datasets(training=0.7, validation=0.2)
#         else:
#             data = self.data.copy()
#             for i in to_delete[::-1]: data.pop(i)
#             return MedicalDataset(data, name=self.name + '.selected')
    
#     def to_subject(self, **kwargs):
#         avouch(self.template is not None)
#         return Subject(**{k: kwargs.get(k, '') for k in self.template.subject.keys()}) \
#               .Image(**{k: kwargs.get(k, '') for k in self.template.image.keys()}) \
#               .SubImage(**{k: kwargs.get(k, '') for k in self.template.subimage.keys()})
    
#     def subimage_infos(self):
#         keys = []
#         for sb, sbd in self.data.items():
#             im, imd = sbd.first()
#             if not isinstance(imd, Dataset): continue
#             n_subimage = im['n_subimage']
#             for l in range(n_subimage):
#                 keys.append(Subject(**sb).Image(n_subimage=n_subimage).to_subimage(l))
#         return keys
    
#     def pair_infos(self):
#         if self.paired_subimage: infos = self.subimage_infos()
#         else: infos = map(Subject, self.data.keys())
#         return infos

#     def randomly_pick_infos(self, n):
#         self.check_data()
#         picked = []
#         infos = self.pair_infos()
#         for _ in range(n): picked.append(infos[random.randint(0, len(infos) - 1)])
#         return picked
    
#     def randomly_pick(self, n):
#         return self.get_batch(self.randomly_pick_infos(n))

#     def split_datasets(self, training = None, validation = None, testing = None):
#         """
#         Split dataset to training, validation and test sets by ratio. e.g. split_dataset(training=0.8, validation = 0.1)
#         """
#         self.check_data()
#         if validation is None and (training is None or testing is None): validation = 0
#         if testing is None and training is None: testing = 0
#         if training is None: training = 1 - validation - testing
#         elif validation is None: validation = 1 - testing - training
#         elif testing is None: testing = 1 - training - validation
#         avouch(training + testing + validation == 1, "Invalid ratios for function 'split_datasets' (Sum is not 1). ")
#         infos = list(self.data.keys())
#         n = len(infos)
#         n_train = int(training * n)
#         n_valid = int(validation * n)
#         random.shuffle(infos)
#         def add_subimage(x):
#             ret = []
#             for s in x:
#                 im, imd = None, None
#                 max_subimage = 0
#                 for img, imgd in self.data[s].items():
#                     if img['n_subimage'] > max_subimage:
#                         max_subimage = img['n_subimage']
#                         im, imd = img, imgd
#                 if not isinstance(imd, Dataset): continue
#                 n_subimage = im['n_subimage']
#                 for l in range(n_subimage):
#                     ret.append(Subject(**s).Image(n_subimage=n_subimage).to_subimage(l))
#             return ret
#         if not self.paired_subimage: add_subimage = lambda x: x
#         self._training_set = add_subimage(infos[:n_train])
#         self._validation_set = add_subimage(infos[n_train:n_train + n_valid])
#         self._testing_set = add_subimage(infos[n_train + n_valid:])
#         return self

#     @alias("train_batch")
#     def training_batch(self, n, **kwargs): return self.batch('training', n, **kwargs)

#     @alias("valid_batch")
#     def validation_batch(self, n, **kwargs): return self.batch('validation', n, **kwargs)

#     @alias("test_batch")
#     def testing_batch(self, n, **kwargs): return self.batch('testing', n, **kwargs)
    
#     def batch(self, stage='training', n_batch=4, shuffle='each_epoch', drop_last=True, none_each_epoch=True, restart=False):
#         self.check_data()
#         if not self._training_set: self.split_datasets(training=0.7, validation=0.2)
#         stage = stage.lower()
#         if stage == 'training': subset = self._training_set
#         elif stage == 'validation': subset = self._validation_set
#         elif stage == 'testing': subset = self._testing_set
#         p = self.batch_pointer[stage]
#         if restart: p = self.batch_pointer[stage] = 0
#         done = False
#         if p < len(subset): 
#             if shuffle: random.shuffle(subset)
#             info_batch = subset[p:p+n]
#             if len(info_batch) == n or not drop_last:
#                 self.batch_pointer[stage] += len(info_batch)
#                 done = True
#         if not done:
#             random.shuffle(subset)
#             self.batch_pointer[stage] = 0
#             if none_each_epoch: return None
#             info_batch = subset[p:p+n]
#             self.batch_pointer[stage] += len(info_batch)
#         return self.get_batch(info_batch)
            
#     def get_batch(self, info_batch):
#         data_arrays = []
#         for info in info_batch:
#             data = self[info.subject]
#             arrays = self.create_batch_func(info.Image(**data.first()[0]), data)
#             if len(arrays) == 0 or arrays is None: return
#             if not data_arrays: data_arrays = [[a] for a in arrays]; continue
#             for i, a in enumerate(arrays):
#                 data_arrays[i].append(a)
#         return tuple((bt.cat(da, {}, crop=True).detach() if da[0].has_batch else bt.stack(da, {}, crop=True).detach()) if isinstance(da[0], bt.Tensor) else da for da in data_arrays)

#     def create_batch(self, func):
#         """
#         Create batch from a subject (e.g. of a patient).
        
#         Args:
#             func (callable): decorated function taking arguments:
#                 subject_info (dict): the info consisting info of the current subject.
#                 group (Dataset Object): the dataset containing the data of this subject.
#                 Returns: torch array without batch dimension, can be a tuple of such arrays. 

#         Examples::
#             >>> @datasetname.create_batch
#             >>> def _(subject_info, group):
#             ...     return group['CT'].to_tensor(), group['MR'].to_tensor() - group['CT'].to_tensor()
#         """
#         self.create_batch_func = func
        



from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "batorch",
    fileinfo = "Optimizers. ",
    requires = "torch"
)

__all__ = """
    CSGD
    CADAM
    Optimization
    train
    test
""".split()

import math
from functools import wraps
from typing import Generator

with __info__:
    import batorch as bt
    from pycamia import alias, avouch, get_environ_vars
    
class _RequiredParameter(object):
    """Singleton class representing a required parameter for an Optimizer."""
    def __repr__(self):
        return "<required parameter>"

required = _RequiredParameter()

class CSGD(bt.optim.Optimizer):
    
    def __init__(self, params, lr=required, momentum=0, dampening=0,
                 weight_decay=0, nesterov=False):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(lr=lr, momentum=momentum, dampening=dampening,
                        weight_decay=weight_decay, nesterov=nesterov)
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        super().__init__(params, defaults)
        self.DGDG = bt.tensor([])
        self.DFDG = bt.tensor([])

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)
            
    def minimize(self, loss, **kwargs):
        self.DGDG = bt.tensor([])
        self.DFDG = bt.tensor([])
        self.zero_grad()
        loss.backward(**self.kwargs)
        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']
            for p in group['params']:
                if p.grad is None:
                    continue
                param_state = self.state[p]
                d_p = p.grad.data
                if weight_decay != 0:
                    d_p.add_(p.data, alpha=weight_decay)
                if momentum != 0:
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = bt.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf
                param_state['loss_grad'] = bt.clone(d_p.data).detach()
                param_state['cons_grad'] = []
        return self
        
    def under_constraint(self, cons):
        self.zero_grad()
        cons.backward()
        innerprods = None
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                param_state = self.state[p]
                if 'cons_grad' not in param_state: param_state['cons_grad'] = []
                param_state['cons_grad'].append(bt.clone(p.grad.data).detach())
                if innerprods is None: innerprods = bt.zeros(len(param_state['cons_grad']) + 1)
                innerprods[0] += (param_state['loss_grad'] * param_state['cons_grad'][-1]).sum().detach()
                innerprods[1:] += (bt.stack(param_state['cons_grad']) * bt.unsqueeze(param_state['cons_grad'][-1])).flatten(1).sum(1).detach()
        self.DFDG = bt.cat((self.DFDG, bt.unsqueeze(innerprods[0])))
        self.DGDG = bt.cat((bt.cat((self.DGDG, innerprods[1:-1].view(1, -1))), innerprods[1:].view(-1, 1)), 1)
        return self

    def step(self, closure=None):
        """Performs a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()
        
        if len(self.DFDG) > 0: lagrange = bt.inverse(self.DGDG) @ self.DFDG.view(-1, 1)
        else: lagrange = []

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                param_state = self.state[p]
                
                d_p = param_state['loss_grad']
                for i, l in enumerate(lagrange):
                    d_p.data.add_(param_state['cons_grad'][i], alpha=-l.item())

                p.data.add_(d_p, alpha=-group['lr'])

        return loss
        
class CADAM(bt.optim.Optimizer):

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0, amsgrad=False):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, amsgrad=amsgrad)
        super().__init__(params, defaults)
        self.cons_values = []
        self.cons_strength = []
        self.DGDG = None
        self.DFDG = None
        self.stepped = False
        self.adjusted = False

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)
            
    def minimize(self, loss, **kwargs):
        self.kwargs = kwargs
        self.zero_grad()
        self.DGDG = None
        self.DFDG = None
        loss.backward(**self.kwargs)
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')
                amsgrad = group['amsgrad']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = bt.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = bt.zeros_like(p.data)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = bt.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                if group['weight_decay'] != 0:
                    grad.add_(p.data, alpha=group['weight_decay'])

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    bt.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = (max_exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                else:
                    denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                    
                state['loss_grad'] = (bt.clone((exp_avg / denom).data).detach() / bias_correction1).special_from_(grad)
                state['cons_grad'] = []
        self.stepped = False
        self.adjusted = False
        self.cons_values = []
        self.cons_strength = []
        return self
        
    def under_constraints(self, cons, strength = 1, **kwargs):
        kwargs.pop('retain_graph', None)
        for i, c in enumerate(cons): self.under_constraint(c, strength=strength, retain_graph = i < len(cons) - 1, **kwargs)
        return self
        
    def under_constraint(self, cons, strength = 1, **kwargs):
        assert 0 <= strength <= 1
        self.zero_grad()
        self.cons_values.append(cons)
        avouch(cons.n_space_dim == 0)
        if cons.has_batch: n_batch = cons.n_batch
        else: n_batch = None
        cons.sum().backward(**kwargs)
        innerprods = None
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None: continue
                param_state = self.state[p]
                if 'cons_grad' not in param_state: param_state['cons_grad'] = []
                param_state['cons_grad'].append(p.grad.detach())
                if innerprods is None:
                    if n_batch is None: innerprods = bt.zeros(len(param_state['cons_grad']) + 1)
                    else: innerprods = bt.zeros([n_batch], len(param_state['cons_grad']) + 1)
                innerprods[..., 0] += (param_state['loss_grad'] * param_state['cons_grad'][-1]).sum().detach()
                DG_x_DG = bt.stack(param_state['cons_grad']) * bt.unsqueeze(param_state['cons_grad'][-1])
                if DG_x_DG.has_batch:
                    if DG_x_DG.ndim > 2: DG_x_DG = DG_x_DG.flatten(2).sum(2)
                    DG_x_DG = DG_x_DG.transpose(0, 1)
                elif DG_x_DG.ndim > 1: DG_x_DG = DG_x_DG.flatten(1).sum(1)
                innerprods[..., 1:] += DG_x_DG.detach()
        if self.DFDG is None: self.DFDG = innerprods[..., :1]
        else: self.DFDG = bt.cat(self.DFDG, innerprods[..., :1], -1)
        if self.DGDG is None: self.DGDG = innerprods[..., 1:].unsqueeze(-1)
        else:
            row_added = bt.cat(self.DGDG, innerprods[..., 1:-1].unsqueeze(-2), -2)
            self.DGDG = bt.cat(row_added, innerprods[..., 1:].unsqueeze(-1), -1)
        self.cons_strength.append(strength)
        return self

    def step(self, closure=None):
        self.stepped = True
        loss = None
        if closure is not None:
            loss = closure()
            
        if self.DFDG is not None:
            self.DGDG += 1e-6 * (bt.det(self.DGDG) <= 1e-6) * bt.eye_as(self.DGDG)
            lagranges = (bt.inverse(self.DGDG) @ self.DFDG.unsqueeze(-1)).squeeze(-1).split(1, -1)
        else: lagranges = []

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                param_state = self.state[p]
                
                d_p = param_state['loss_grad']
                for i, l in enumerate(lagranges):
                    d_p.data.add_(l * param_state['cons_grad'][i], alpha = -self.cons_strength[i])

                p.data.add_(d_p, alpha=-group['lr'])

        return loss
        
    def adjust(self, *G, to_zero = False):
        if not self.stepped: raise TypeError("step not called")

        if len(G) > 0: eps = (bt.inverse(self.DGDG) @ (bt.stack(G, -1).unsqueeze(-1) - bt.stack(self.cons_values, -1).unsqueeze(-1) * int(not to_zero))).squeeze(-1)
        else: eps = []

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None: continue
                param_state = self.state[p]
                for i, e in enumerate(eps.split(1, -1)):
                    p.data.add_(-e * self.cons_strength[i] * param_state['cons_grad'][i])

def strify(x):
    if isinstance(x, bt.torch.Tensor):
        if x.numel() == 1: return x.item()
        else: return x.tolist()
    return x

class Optimization:
    
    def __init__(self, optimizer, params_or_model, dataset = None, lr = 1e-3, **kwargs):
        if isinstance(params_or_model, bt.nn.Module):
            init_params = params_or_model.parameters()
            self.model = params_or_model
            self.dataset = dataset
        else:
            init_params = params_or_model
            self.model = None
            self.dataset = dataset
        self.hyper = dict(lr = lr if callable(lr) else lambda _: lr)
        self.hyper.update({k: v if callable(v) else lambda _: v for k, v in kwargs.items()})
        if isinstance(init_params, (Generator, tuple)): init_params = list(init_params)
        if not isinstance(init_params, list): init_params = [init_params]
        if not isinstance(init_params[0], dict) or 'params' not in init_params[0]:
            init_params = [dict(params = x, **{k: v(0) for k, v in self.hyper.items()}) for x in init_params]
        self._optimizer = optimizer(init_params)
        self._index = 1
        self.loss = []
        self.loss_record = {}

    def optimizer(self, i):
        # state_dict = self._optimizer.state_dict()
        # # for param_state_dict in state_dict['param_groups']:
        # #     for k in self.hyper:
        # #         param_state_dict[k] = self.hyper[k](i)
        # #         print(f"{k}: {self.hyper[k](i)}")
        # self._optimizer.load_state_dict(state_dict)
        for param_group in self._optimizer.param_groups:
            for k in self.hyper:
                param_group[k] = self.hyper[k](i)
        return self._optimizer
    
    def new_iter(self):
        optimizer = self.optimizer(self._index)
        self._index += 1
        return optimizer
            
    def minimize(self, loss, **kwargs):
        """Minimize the loss function in the optimizer.

        Args:
            loss (bt.Tensor or callable): The loss funtion, a batorch tensor with graph
                or a function taking arguments (model, batch_data).
            Note: batch_data is obtained by calling method 'training_batch(n_batch)' in optimizer.dataset.
        """
        if callable(loss): self.loss.append((1, loss)); return
        optimizer = self.optimizer(self._index)
        optimizer.zero_grad()
        loss.backward(**kwargs)
        optimizer.step()
        self._index += 1

    def maximize(self, loss, **kwargs):
        """Maximize the loss function in the optimizer.

        Args:
            loss (bt.Tensor or callable): The loss funtion, a batorch tensor with graph
                or a function taking arguments (model, batch_data).
            Note: batch_data is obtained by calling method 'training_batch(n_batch)' in optimizer.dataset.
        """
        if callable(loss): self.loss.append((-1, loss)); return
        optimizer = self.optimizer(self._index)
        optimizer.zero_grad()
        (-loss).backward(**kwargs)
        optimizer.step()
        self._index += 1

    @alias("optimize")
    def train(self, max_epoch=10, n_epoch_valid=1, n_batch=4):
        """
        Optimize the function [train the model] by this optimizer. 
        Warning: Requires Dataset element in package `micomputing`. 

        Args:
            max_epoch (int): The number of times we scan the dataset, i.e. the number of epochs. Defaults to `10`.
            n_epoch_valid (int): Validate every n_epoch_valid epochs. Defaults to `1`.
            n_batch (int): Batch size for training and validation. Defaults to `4`.
        """
        avouch(self.dataset is not None and self.model is not None, 
               "Model and dataset needed for training an optimization problem. ")
        epoch_strlen = len(str(max_epoch))
        for i_epoch in range(1, max_epoch + 1):
            epoch = f'%0{epoch_strlen}d'%i_epoch
            i = 1
            while True:
                self.model.train()
                batch = self.dataset.training_batch(n_batch)
                if batch is None: break
                loss = self.compute_loss(self.model, batch)
                self.minimize(loss)
                print(f"[TRAIN] epoch = {epoch}, iteration = {'%04d'%i}, loss = {strify(loss)}")
                i += 1
            if i_epoch % n_epoch_valid == 0:
                self.model.eval()
                batch = self.dataset.validation_batch(n_batch, restart = True)
                if batch is None: break
                loss = self.compute_loss(self.model, batch)
                print(f"[VALID] epoch = {epoch}, loss = {strify(loss)}")

    def compute_loss(self, model, batch):
        total_loss = bt.tensor(0)
        for d, l in self.loss:
            l_value = l(model, batch)
            self.loss_record[l.__name__] = l_value
            total_loss += d * l_value
        return total_loss

    def get_loss(self, name):
        return self.loss_record.get(name, None)

    def __getitem__(self, i): return self.optimizer(i)

def train(dataset, max_epoch=10, n_epoch_valid=1, n_batch=4, iter_in_epoch=None, valid_info=None):
    """
    Train by dataset. 
    Warning: Requires `Dataset` in package `micomputing`. 
    
    Note: Please use a function for loss computation (see example). It takes index and data batch. 
        index is a tuple of 3, including epoch index, iteration in the current epoch and the total iterations. 
        The latter two are 0s for validation.

    Args:
        dataset (Dataset): The dataset structure.
        max_epoch (int): The number of times we scan the dataset, i.e. the number of epochs. Defaults to `10`.
        n_epoch_valid (int): Validate every n_epoch_valid epochs. Defaults to `1`.
        n_batch (int): Batch size for training and validation. Defaults to `4`.
        valid_info (None or list): The manually specified batch to validate.
    
    Examples::
        >>> @train(dataset)
        ... def loss(index, batch_data):
        ...     stage, epoch, iteration, total_iter = index
        ...     # stage: 'training' / 'validation'; iteration: i_iter in current epoch
        ...     model1.train()
        ...     loss1 = loss_func(batch_data)
        ...     opt1.minimize(loss1)
        ...     if total_iter % 100 == 0:
        ...         loss2 = loss_func(batch_data)
        ...         opt2.maximize(loss2)
        ... return {'loss1': loss1, 'loss2': loss2}
    """
    def decorator(func):
        k = 1
        epoch_strlen = len(str(max_epoch))
        for i_epoch in range(1, max_epoch + 1):
            epoch = f'%0{epoch_strlen}d'%i_epoch
            i = 1
            while True:
                batch = dataset.training_batch(n_batch)
                if batch is None: break
                if iter_in_epoch is not None and i > iter_in_epoch: break
                results = func(('training', i_epoch, i, k), batch)
                if not isinstance(results, dict): results = {'loss': results}
                print(f"[TRAIN] epoch = {epoch}, iteration = {'%04d'%i}, " + ', '.join([f"{k} = {strify(v)}" for k, v in results.items()]))
                i += 1; k += 1
            if i_epoch % n_epoch_valid == 0:
                if valid_info is not None: batch = dataset.get_batch(valid_info)
                else: batch = dataset.validation_batch(n_batch, restart = True)
                if batch is None: break
                results = func(('validation', i_epoch, 0, 0), batch)
                if not isinstance(results, dict): results = {'loss': results}
                print(f"[VALID] epoch = {epoch}, " + ', '.join([f"{k} = {strify(v)}" for k, v in results.items()]))
    return decorator

def test(dataset, n_batch=4, test_info=None):
    """
    Test by dataset. 
    Warning: Requires `Dataset` in package `micomputing`. 
    
    Note: Please use a function for loss computation (see example). It takes index and data batch. 
        index is a number, indicating the iteration of batch. It is `0` when `test_info` is given.

    Args:
        dataset (Dataset): The dataset structure.
        n_batch (int): Batch size for testing. Defaults to `4`.
        test_info (None or list): The manually specified batch to test.
    
    Examples::
        >>> @test(dataset)
        ... def loss(index, batch_data):
        ...     i_batch = index
        ...     model1.eval()
        ...     loss = loss_func(batch_data)
        ...     plt.plot(batch_data[0])
        ... return {'loss': loss}
    """
    def decorator(func):
        if test_info is not None:
            batch = dataset.get_batch(test_info)
            results = func(0, batch)
            if not isinstance(results, dict): results = {'loss': results}
            print("[TEST] " + ', '.join([f"{k} = {strify(v)}" for k, v in results.items()]))
        else:
            i = 1
            while True:
                batch = dataset.testing_batch(n_batch)
                if batch is None: break
                results = func(i, batch)
                if not isinstance(results, dict): results = {'loss': results}
                print(f"[TEST] ibatch = {i}, iteration = {'%04d'%i}, " + ', '.join([f"{k} = {strify(v)}" for k, v in results.items()]))
                i += 1
    return decorator
